#!/usr/bin/python
## -*- coding: utf-8 -*-
## (C) 2007, 2008, 2013-2015 Muthiah Annamalai
## Licensed under GPL Version 3
##
## This module is the runtime elements for the exprs language.
## It contains classes 
## 

from .errors import RuntimeException
import keyword
import sys
import time
import traceback
from .profile import Profiler

PYTHON3 = (sys.version[0] == '3')
if PYTHON3:
    str = str

# customize I/O like functions for external GUI
class tamilCustomFunction(object):
    _instance = None
    def __init__(self):
        object.__init__(self)
        self.input_fcn = input
    @staticmethod
    def get_instance():
        if not tamilCustomFunction._instance:
            tamilCustomFunction._instance = tamilCustomFunction()
        return tamilCustomFunction._instance
    
    @staticmethod
    def input(args):
        return tamilCustomFunction.get_instance().input_fcn(args)
    
    @staticmethod
    def reset():
        if PYTHON3:
            tamilCustomFunction.get_instance().input_fcn = input
        else:
            tamilCustomFunction.get_instance().input_fcn = input
    
    @staticmethod
    def set(newfcn):
        tamilCustomFunction.get_instance().input_fcn = newfcn
    
## RUNTIME / LIBRARY etc
class DebugUtils:
    def __init__(self,dbg):
        self.debug = dbg

    def dbg_msg(self, msg):
        """ handy to print debug messages """
        if ( self.debug ):
            print(u"## ",msg)
        return

class BuiltinFunction:
    """ a templatized way to invoke bulitin functions."""
    def __init__(self,fn,name,padic=1,dbg = False):
        self.fn = fn
        self.name = name
        self.use_adicity = True
        self.padic = padic
        self.debug = dbg
        self.aslist = False
        
    def __str__(self):
        return u"[BuiltinFunction[%s,nargs=%d]]"%(self.name,self.padic)
    
    def evaluate(self,env):
        try:
            return self.do_evaluate(env)
        except AssertionError as assert_excep:
            env.unroll_stack()
            raise RuntimeException( str(assert_excep) + u' Assertion failed!')
        except Exception as excep:
            env.unroll_stack()            
            print(u"failed dispatching function ",str(self),u"with exception",str(excep))
            if ( self.debug ):
                traceback.print_tb(sys.exc_info()[2])
            raise excep
        
    def do_evaluate(self,env):
        ## push stuff into the call-stack
        env.call_function(self.name)
        
        ## check arguments match, otherwise raise error
        args = env.get_args()#.get_list()
        env.set_local({})

        if ( self.use_adicity and len(args) < self.padic ):
            raise RuntimeException("Too few args to bulitin function " + self.name)

        # keep evaluating for as many evaluate object-methods are available
        # because Python libraries don't recognize tamil AST objects.
        while( any(filter(lambda x: hasattr(x,'evaluate'),args)) ):            
            args_ = []
            for a in args:
                if hasattr(a,'evaluate'):
                    a = a.evaluate(env)
                args_.append(a)
            args = args_
        
        if ( self.use_adicity ):
            if ( self.debug ): print(self.fn, args, self.padic)
            rval = self.fn(*args)
        else:
            try:                
                if ( self.aslist ):
                    rval = self.fn(*[args])
                else:
                    rval = self.fn(*args)
            except AssertionError as assert_excep:
                raise assert_excep
            except Exception as excep:
                #print u"catch exception", excep
                raise RuntimeException( str(excep) )
        env.clear_call(copyvars= self.name in ["execute","eval",u'இயக்கு',u'மதிப்பீடு'])
        ## pop stuff into the call-stack
        env.return_function(self.name)
        return  rval

class BlindBuiltins(BuiltinFunction):
    """ also blindly include all functions from the 
        os, sys, curses.ascii, math etc. donot check arguments 
        here.    """
    def __init__(self,fn,name,dbg = False,aslist=False,padic=-1):
        BuiltinFunction.__init__(self,fn,name,padic,dbg)
        self.use_adicity = False
        self.aslist = aslist
        
    def __str__(self):
        return BuiltinFunction.__str__(self)

class CountDownTimer(object):
    def __init__(self):
        object.__init__(self)
        self._engaged = False
        self._safeModeTimeout = 13  # seconds of wall clock timeout for tamil codes in SAFE MODE
        self._startTime = 0

    # all times are wall clock times
    def is_timeout(self):
        if not self._engaged:
            return False
        now = time.time()
        rval = (now - self._startTime) > self._safeModeTimeout
        if rval:
            self._engaged = False
        return rval

    def record_start_time(self):
        self._startTime = time.time()
        self._engaged = True
        return

    # update safe mode timeout
    def set_timeout(self, seconds):
        assert seconds > 0
        self._safeModeTimeout = seconds


# <<Side-Effects>>: computation is side-effect of programming.
class Environment(CountDownTimer):
    """ used to manage the side-effects of an interpreter """
    def __init__(self, call_stack, function_map, builtin_map, dbg = False, MAX_REC_DEPTH  = 128 ):
        CountDownTimer.__init__(self)
        self.do_profiling = False
        self.profiler = None
        self.max_recursion_depth = MAX_REC_DEPTH #keep it smaller than Python stack
        self.call_stack = call_stack#list
        self.function_map = function_map#dicts
        self.builtin_map = builtin_map#dicts
        self.local_vars = []#list of dicts.
        self.arg_stack = []#list of lists
        self.ret_stack = [] #list
        self.debug = dbg #use to turn debugging on.
        self.readonly_global_vars = {'True':True, 'False':False,u"மெய்":True, u"பொய்":False} #dict of global vars
        self.clear_break_return_continue()

    def validate_timer(self):
        # time to be reset everytime the interpreter calls evaluate
        if self.is_timeout():
            raise RuntimeException("Evaluation of your program is taking too long!\n"
                                   "\tTimeout exceeded %g (seconds)"%self._safeModeTimeout)
        return

    def __del__(self):
        if (self.debug): print(u"deleting environment")
        
    def enable_profiling(self):
        self.validate_timer()
        self.do_profiling =True
        self.profiler = Profiler()
        return
    
    def is_profiling(self):
        self.validate_timer()
        return self.do_profiling
    
    def disable_profiling(self):
        self.validate_timer()
        self.do_profiling = False
        
    def report_profiling(self):
        self.validate_timer()
        self.disable_profiling()
        self.profiler.report_stats()
        
    def reset_profiling(self):
        self.validate_timer()
        self.disable_profiling()
        self.profiler = None
        
    def unroll_stack(self):
        self.validate_timer()
        # we skip BOS since it is a __toplevel__
        if (self.debug): print("clearing locals")
        if len(self.call_stack) > 0:
            if self.call_stack[-1][0] != "__toplevel__":
                self.clear_call()
        
        for i in range(1,len(self.call_stack)):
            print(u"%sError in function '%s'%s:"%(u" "*(i-1),self.call_stack[i][0],self.call_stack[i][1]))
        while len(self.call_stack) > 0:
            tos,_ = self.call_stack.pop()
            #print(u"Error in location %s"%str(tos))
        return
    
    def get_break_return(self):
        self.validate_timer()
        """ get if break or return was set for use in loops """
        val = self.Break or self.Return
        return val

    def clear_break(self):
        self.validate_timer()
        """ reset after a break statement """
        self.Break = False
        return False

    def clear_continue(self):
        self.validate_timer()
        """ reset after a continue statement """
        self.Continue = False
        return False

    def set_break(self):
        self.validate_timer()
        """ execute a continue statement """
        self.Break = True
        return True

    def set_continue(self):
        self.validate_timer()
        """ execute a continue statement """
        self.Continue = True
        return True
    
    def break_return_continue(self):
        self.validate_timer()
        val = ( self.Break or 
                 self.Return or 
                 self.Continue )
        ## must clear continue flag right away.
        if ( self.Continue ):
            self.Continue = False
        return val
    
    def dbg_msg(self, msg):
        """ handy to print debug messages """
        if ( self.debug ):
            print(msg)
        return
    
    def __str__(self):
        if (self.debug):
            return repr(self)
        return u"<env>"
    
    def __repr__(self):
        retval = u"CallStack =>"+str(self.call_stack) + u"\n" \
            u"LocalVars =>"+ str(self.local_vars) + u"\n" \
            u"ArgStack =>"+ str(self.arg_stack) + u"\n" 
        return retval

    def set_retval( self, rval ):
        self.validate_timer()
        self.ret_stack.append( rval )
        self.Return = True
        return
    
    def get_retval( self ):
        self.validate_timer()
        rval = None
        if ( len(self.ret_stack ) >= 1  ):
            rval = self.ret_stack.pop()
        return rval

    def clear_break_return_continue(self):
        self.validate_timer()
        self.Break = False
        self.Return = False
        self.Continue = False

    def clear_call(self,copyvars=False):
        self.validate_timer()
        """ utility to cleanup the stacks etc.. """
        self.clear_local( copyvars ) 
        self.clear_args ( )
        self.clear_break_return_continue()
        
    def clear_args(self):
        self.validate_timer()
        """ cleanup the stack """
        self.arg_stack.pop()
        return

    def get_args(self):
        self.validate_timer()
        """ manage a global argument stack """
        return self.arg_stack[-1]

    def set_args(self,val):
        self.validate_timer()
        """ manage a global argument stack """
        self.dbg_msg( "setting args " + str( val ) )
        return self.arg_stack.append(val)
    
    def set_local(self, vars):
        self.validate_timer()
        self.local_vars.append(vars)
        self.dbg_msg( "setting locals " + str( vars ) )
        self.clear_break_return_continue()
        return
    
    def clear_local(self,copyvars=False):
        self.validate_timer()
        prev_values = self.local_vars.pop()
        if copyvars:
            if len(self.local_vars) < 1:
                self.local_vars.append({})
            self.local_vars[-1].update(prev_values)
        return
    
    def has_id(self, idee):
        """ check various 'scopes' for ID variable """
        self.validate_timer()
        rval = False
        if idee in self.readonly_global_vars:
            return True
        if ( len( self.local_vars ) == 0 ):
            return False
        variables = self.local_vars[-1]
        rval = idee in variables
        return rval

    def set_id(self, idee, val, global_id = False):
        """ someday do global_id """
        self.validate_timer()
        if idee in self.readonly_global_vars:
            raise Exception(u"Error: Attempt to reassign constant %s"%idee)
        if ( len(self.local_vars) > 0 ):
            d=self.local_vars[-1]
        else:
            d=dict()
            self.local_vars.append(d)
        #if not (idee in d) and (idee in self.global_vars):
        #    self.global_vars[idee] = val
        # else:
        d[idee]=val
        self.dbg_msg("set_id: " + str(idee) +" = "+str(val))
        return

    def get_id(self, idee):
        self.validate_timer()
        val = None
        if idee in self.readonly_global_vars:
            return self.readonly_global_vars[idee]
        if not self.has_id(idee):
            note = ''
            if idee in keyword.kwlist:
                note = 'Did you possibly confuse the Python english keyword %s for tamil keyword ?'%idee
            raise RuntimeException("Identifier %s not found"%idee)
        variables = self.local_vars[-1]
        val = variables[idee]
        self.dbg_msg("get_id: val = "+str(val))
        return val

    def call_function(self, fn,callsite=u""):
        """ set call stack, used in function calls. Also check overflow"""
        self.validate_timer()
        if ( len(self.call_stack) >= self.max_recursion_depth ):
            raise RuntimeException( "Maximum recursion depth [ " + 
                                    str(self.max_recursion_depth) + 
                                    " ] exceeded; stack overflow." )
        self.dbg_msg(u"calling function"+str(fn)+callsite)
        if ( self.is_profiling() ):
            self.profiler.add_function( fn )
        self.call_stack.append( (fn,callsite) ) 
    
    def return_function(self, fn):
        self.validate_timer()
        va,_ = self.call_stack.pop( )
        if ( fn != va ):
            raise RuntimeException("function %s doesnt match Top-of-Stack"%fn)
        if ( self.is_profiling() ):
            self.profiler.update_function( va )
        return va
    
    def has_function(self, fn):
        self.validate_timer()
        if ( fn in self.builtin_map ):
            return True
        if ( fn in self.function_map ):
            return True
        return False
    
    def get_function(self, fn):
        self.validate_timer()
        if not self.has_function(fn):
            raise RuntimeException("undefined function: "+fn)

        if ( fn in self.builtin_map ):
            return self.builtin_map[fn]
        
        if ( fn in self.function_map ):
            return self.function_map[fn] 
        
        raise RuntimeException("Environment error on fetching function "+fn)



#!/usr/bin/python
##
## (C) 2007, 2008, 2015 Muthiah Annamalai
## Licensed under GPL Version 3
##
## This module is the scanner for the exprs language.
## It contains classes Token, Lexeme, and  Lex.
## 

from .errors import ScannerException
import codecs
import re
import sys
import tamil

PYTHON3 = (sys.version[0] == '3')
if PYTHON3:
    str = str

## SCANNER
class Token:
    token_types=[u"EOF",u"ID",u"NUMBER",u"PRINT",u"+",u"-",
                 u"*",u"/",u"(",u")",u",",
                 u"=",u"END",u"DEF",u"RETURN",u"IF",u"ELSEIF",u"ELSE",
                 u"DO",u"WHILE",u"FOR",u"STRING",u">",u"<",u">=",u"<=",u"!=",
                 u"==",u"[",u"]",u"^",u"%",u"BREAK",u"CONTINUE",u"SWITCH",
                 u"CASE",u"OTHERWISE",u"&&",u"||",u"&",u"|",u"!",
                 u"{",u"}",u":",u"<<",u">>",u"~"]
    
    @staticmethod
    def is_string(kind):
        """ predicate to check if @kind token is a string """
        return Token.get_name(kind) == u"STRING"
    
    @staticmethod
    def is_number(kind):
        """ predicate to check if @kind token is a number """
        return Token.get_name(kind) == u"NUMBER"
    
    @staticmethod
    def is_id(kind):
        """ predicate to check if @kind token is an identifier """
        return Token.get_name(kind) == u"ID"

    @staticmethod    
    def is_keyword(kind):
        """ predicate to identify if @kind token is a keyword """
        if (  Token.get_name(kind) in ["END","FOR","WHILE","DO","IF","ELSEIF","ELSE","DEF","SWITCH","CASE","OTHERWISE"] ):
            return True
        return False
    
    @staticmethod
    def get_name(kind):
        """ used in reporting errors in match() on parsing stage """
        if ( kind >= 0 and kind < len(Token.token_types) ):
            return Token.token_types[kind]
        raise ScannerException(u"Index out of bounds. Unknown token "+str(kind))
        return None    
    
    EOF = 0
    ID = 1
    NUMBER = 2
    PRINT = 3
    PLUS = 4
    MINUS = 5
    PROD = 6
    DIV = 7
    LPAREN = 8
    RPAREN = 9
    COMMA = 10
    EQUALS = 11
    END = 12
    DEF = 13
    RETURN = 14
    IF = 15
    ELSEIF = 16
    ELSE   = 17
    DO = 18
    WHILE = 19
    FOR = 20
    STRING = 21
    GT = 22
    LT = 23
    GTEQ = 24
    LTEQ = 25
    NEQ = 26
    EQUALITY = 27
    LSQRBRACE = 28
    RSQRBRACE = 29
    EXP = 30
    MOD = 31
    BREAK = 32
    CONTINUE = 33
    SWITCH = 34
    CASE = 35
    OTHERWISE = 36
    LOGICAL_AND = 37
    LOGICAL_OR = 38
    BITWISE_AND = 39
    BITWISE_OR = 40
    LOGICAL_NOT = 41
    LCURLBRACE = 42
    RCURLBRACE = 43
    COLON = 44
    BITWISE_LSHIFT = 45
    BITWISE_RSHIFT = 46
    BITWISE_COMPLEMENT = 47
    
    ## operator classifications
    UNARYOP = [LOGICAL_NOT,BITWISE_COMPLEMENT]
    ADDSUB = [PLUS, MINUS]
    MULDIV = [PROD,DIV]
    COMPARE = [ GT, LT, GTEQ, LTEQ, NEQ, EQUALITY ]
    EXPMOD = [EXP, MOD]
    BITWISE_AND_LOGICAL = [LOGICAL_AND, LOGICAL_OR, BITWISE_AND, BITWISE_OR, BITWISE_LSHIFT, BITWISE_RSHIFT]
    BINOP = []
    for i in [ADDSUB, MULDIV, COMPARE, EXPMOD, BITWISE_AND_LOGICAL]:
        BINOP.extend( i )
    ASSIGNOP = [EQUALS]

class Lexeme:
    def __init__(self,val,kind,fname=""):
        self.val = val
        self.kind = kind
        self.line = -1
        self.col = -1
        self.fname = fname

    def set_file_name(self,file_name):
        self.fname = file_name
        return True
        
    def get_line_col(self):
        return [self.line, self.col]

    def set_line_col(self, lc):
        [l, c] = lc
        self.line = l
        self.col = c
        
    def __str__(self):
        return u" %s Line=%d, Col=%d in File %s "% \
            (str(self.val),self.line,self.col,self.fname)

class DummyFile:
    """ wrap a bunch of string data in a file interface """
    def __init__(self,data):        
        self.data = data;        
    
    def close(self):
        pass
    
    def readlines(self):
        return self.data.split(u"\n")

class Lex(object):
    """ Lexer automatically starts lexing on init.
    Maybe use some Library module? """

    def __init__(self,fname=None,dbg=False,encoding="utf-8"):
        object.__init__(self)
        self.encoding=encoding.lower()
        self.stdin_mode = False
        self.debug = dbg
        self.converted_data = None
        self.comments = {} #comments dict indexed by line number with comments present as string value
        if ( isinstance(fname,str) or isinstance(fname,str) ):
            self.fname = fname
            if self.encoding == "utf-8":
                if( self.debug ): print(u"-> opening file %s",fname)
                self.File = codecs.open(fname,"r","utf-8")
            elif self.encoding == "tscii":
                # use open-tamil libraries to translate the source file
                self.File = open(fname,"r")
                self.converted_data = tamil.tscii.convert_to_str( self.File.read() )
                if (self.debug ):
                    print("########## TSCII CONVERSION  TO str DONE ##########")
                    print("######################### %d letters found #################"%len(self.converted_data))
            else:
                raise Exception("Unknown encoding %s used for file %s"%(encoding,fname))    
        elif ( isinstance(fname,list) ):
            """ specify, fname = ["contents of program as a string"] """
            self.fname = u"<DUMMYFILE>"
            self.File = DummyFile( fname[0] );
        else:
            self.fname = u"<STDIN>"
            self.stdin_mode = True
        if ( self.debug ): 
            print(u"post file open")
        ##actual col = idx - col_idx
        self.line = 1 
        self.col_idx = 0 
        ##contains lexeme's in reverse order
        ##for popping is elegant.
        self.tokens=[]
        self.spc=re.compile(u"\s+")
        self.newlines=re.compile(u"\n+")
        self.unary_binary_ops = \
            [ '+','-','=','*','/','>','<','%','^','!=','!','&&','||','|','&','!','>>','<<','~']
        ## need to be the last on init & only for files
        if ( not self.stdin_mode  ): self.tokenize()
        ## REPL loop can call tokenize_string whenever it
        ## desires so.

    def reset(self):
        """ reset the lexer """
        if (self.debug): print(u"Dumping out "+ str(len(self.tokens))+"Lexemes ")
        self.tokens = []
    
    def __repr__(self):
        if ( self.debug ):
            for idx in range(0,len(self.tokens)):
                print(u"%d] %s"%(idx,repr(self.tokens.pop())))
        return u""

    def get_lexeme(self,chunks , pos):        
        if chunks == None:
            return None

        if chunks == u"print":
            tval=Lexeme(chunks,Token.PRINT)
        elif chunks == u"if":
            tval = Lexeme( chunks, Token.IF )
        elif chunks == u"elseif":
            tval = Lexeme( chunks, Token.ELSEIF )
        elif chunks == u"else":
            tval = Lexeme( chunks, Token.ELSE )
        elif chunks == u"for":
            tval = Lexeme( chunks, Token.FOR )
        elif chunks == u"while":
            tval = Lexeme( chunks, Token.WHILE )
        elif chunks == u"do":
            tval = Lexeme( chunks, Token.DO )
        elif chunks == u"return":
            tval=Lexeme(chunks,Token.RETURN)
        elif chunks == u"end":
            tval=Lexeme(chunks,Token.END)
        elif chunks == u"def":
            tval=Lexeme(chunks,Token.DEF)
        elif chunks == u"continue":
            tval=Lexeme(chunks,Token.CONTINUE)
        elif chunks == u"break":
            tval=Lexeme(chunks,Token.BREAK)
        elif chunks == u"=":
            tval=Lexeme(chunks,Token.EQUALS)
        elif chunks == u"-":
            tval=Lexeme(chunks,Token.MINUS)
        elif chunks == u"+":
            tval=Lexeme(chunks,Token.PLUS)
        elif chunks == u">":
            tval=Lexeme(chunks,Token.GT)
        elif chunks == u"<":
            tval=Lexeme(chunks,Token.LT)
        elif chunks == u">=":
            tval=Lexeme(chunks,Token.GTEQ)
        elif chunks == u"<=":
            tval=Lexeme(chunks,Token.LTEQ)
        elif chunks == u"==":
            tval=Lexeme(chunks,Token.EQUALITY)
        elif chunks == u"!=":
            tval=Lexeme(chunks,Token.NEQ)
        elif chunks == u"*":
            tval=Lexeme(chunks,Token.PROD)
        elif chunks == u"/":
            tval=Lexeme(chunks,Token.DIV)
        elif chunks == u",":
            tval=Lexeme(chunks,Token.COMMA)
        elif chunks == u"(":
            tval=Lexeme(chunks,Token.LPAREN)
        elif chunks == u")":
            tval=Lexeme(chunks,Token.RPAREN)
        elif chunks == u"[":
            tval=Lexeme(chunks,Token.LSQRBRACE)
        elif chunks == u"]":
            tval=Lexeme(chunks,Token.RSQRBRACE)
        elif chunks == u"{":
            tval=Lexeme(chunks,Token.LCURLBRACE)
        elif chunks == u"}":
            tval=Lexeme(chunks,Token.RCURLBRACE)
        elif chunks == u":":
            tval=Lexeme(chunks,Token.COLON)
        elif chunks == u"%":
            tval=Lexeme(chunks,Token.MOD)
        elif chunks == u"^":
            tval=Lexeme(chunks,Token.EXP)
        elif chunks == u"&&":
            tval=Lexeme(chunks,Token.LOGICAL_AND)
        elif chunks == u"&":
            tval=Lexeme(chunks,Token.BITWISE_AND)
        elif chunks == u"||":
            tval=Lexeme(chunks,Token.LOGICAL_OR)
        elif chunks == u"<<":
            tval=Lexeme(chunks,Token.BITWISE_LSHIFT)
        elif chunks == u">>":
            tval=Lexeme(chunks,Token.BITWISE_RSHIFT)
        elif chunks == u"~":
            tval=Lexeme(chunks,Token.BITWISE_COMPLEMENT)
        elif chunks == u"|":
            tval=Lexeme(chunks,Token.BITWISE_OR)
        elif ( chunks[0] == u"\"" and chunks[-1] == u"\"" ):
            tval = Lexeme( chunks[1:-1], Token.STRING )
        elif chunks[0].isdigit() or chunks[0]==u'+' or chunks[0]==u'-':
            # deduce a float or integer
            if ( chunks.find(u'.') >= 0 or chunks.find(u'e') >= 0 or chunks.find(u'E') >= 0 ):
                tval=Lexeme(float(chunks),Token.NUMBER)
            else:
                tval=Lexeme(int(chunks),Token.NUMBER)
        elif chunks[0].isalpha():
            tval=Lexeme(chunks,Token.ID)
        elif chunks in  ['\r','\n','\r\n']:
            return None
        else:
            print("==>",chunks,"<==")
            scanner_exception = u"Lexical error: " + str(chunks) + u" at Line , Col "+str(self.get_line_col( pos )) +u" in file "+self.fname;
            raise ScannerException( scanner_exception )
        
        [l,c]=self.get_line_col( pos )
        tval.set_line_col( [l,c] )
        tval.set_file_name( self.fname )
        self.tokens.append( tval )
        return l
    
    def update_line_col(self, pos ):
        ## pos is in current token stream
        self.line = self.line + 1
        self.col_idx = pos 

    def get_line_col( self , pos ):
        ## pos is in current token stream
        return [ self.line, pos -  self.col_idx  ]

    def set_line_col(self, lc):
        [l, c] = lc
        self.line = l
        self.col_idx = c

    def tokenize(self,data=None):
        """ do hard-work of tokenizing and
        put Lexemes into the tokens[] Q """
        if ( self.debug ): print( u"tokenize")
        if ( self.stdin_mode ):
            if ( self.debug ): print(self.tokens)
            ## cleanup the Q for stdin_mode of any EOF that can remain.
            if ( len(self.tokens) != 0 ):
                self.match( Token.EOF )
            if( len(self.tokens) != 0 ):
                raise ScannerException("Lexer: token Q has previous session tokens ")
            self.tokens = list()
        else:
            if self.encoding == "utf-8":
                data = u"".join(self.File.readlines())
            elif self.encoding == "tscii":
                data = self.converted_data
            else:
                assert False
        idx = 0 
        tok_start_idx = 0

        while ( idx < len( data ) ):
            if ( self.debug ): print(idx, data[idx])
            c = data[idx]
            
            if  ( c == ' 'or c == '\t' or c == '\n' ):
                if ( c == '\n' ):
                    ##actual col = idx - col_idx
                    self.update_line_col(idx)
                idx = idx + 1
            elif ( c == '#' ):
                ## single line skip comments like Python/Octave
                while ( idx < len( data ) and data[idx] !='\n' ):
                    idx = idx + 1                    
            elif ( c.isdigit() or c == '+' or c == '-'  ):
                num = c
                tok_start_idx = idx
                idx = idx + 1
                ## FIXME: this prevents you from +.xyz, or -.xyz use 0.xyz 
                ## instead. also may throw an error if we exceed 
                ## buffer-length.                
                if ( c in ['+','-']  and ( idx < len( data ) ) 
                     and not data[idx].isdigit() ):
                    self.get_lexeme( c , idx )
                    continue
                while ( ( idx < len( data) )
                            and ( data[idx].isdigit() or data[idx] == '.') ):
                    num = num + data[idx]
                    idx = idx + 1
                self.get_lexeme( num , tok_start_idx  )
            elif ( c == "\"" ):
                tok_start_idx = idx 
                s = c
                idx = idx + 1
                while ( idx < len( data ) and
                         ( data[idx] != '\"' ) ):
                    s = s + data[idx]
                    if ( data[idx] == '\\' ):
                        idx = idx + 1
                    idx  = idx + 1
                s = s+data[idx]
                idx  = idx + 1
                self.get_lexeme( s , tok_start_idx )
            elif ( c.isalpha() ):
                tok_start_idx = idx 
                s = c
                idx = idx + 1
                while ( ( idx < len( data ) )
                            and ( data[idx].isalpha() or data[idx].isdigit() )
                                  or data[idx] in [ "\"", "_" ] ):
                    s = s + data[idx]
                    idx = idx + 1
                self.get_lexeme( s , tok_start_idx )
            elif ( c in self.unary_binary_ops ):
                tok_start_idx = idx 
                if ( len(data) > ( 1 + idx  ) 
                     and data[idx+1] in ['=','|','&','>','<']  ):
                    c = c +data[idx+1]
                    idx = idx + 1
                self.get_lexeme(  c , tok_start_idx )
                idx = idx + 1
            else:
                tok_start_idx = idx 
                idx = idx + 1
                self.get_lexeme( c , tok_start_idx )
        
        tok_start_idx = idx 

        ## close the file if not stdin_mode
        if ( not self.stdin_mode ): self.File.close()

        ## and manually add an EOF statement.
        eof_tok = Lexeme(u"",Token.EOF )
        eof_tok.set_line_col( self.get_line_col( tok_start_idx ) )
        self.tokens.append( eof_tok )

        self.tokens.reverse()
        return 
        
    def dump_tokens(self):
        print(u" \n".join([str(self.tokens[i]) for i in range(len(self.tokens)-1,-1,-1)]))
        return

    def dequeue(self):
        """ remove Lexeme from the head of Q"""
        ##print "~ ~ match ",repr(self.tokens[-1])
        return self.tokens.pop()
    
    def end_of_tokens(self):
        return (len(self.tokens) == 0)  or (self.peek().kind == Token.EOF )

    def peek(self):
        """ remove Lexeme from the head of Q""" 
        if len(self.tokens) == 0:
            raise ScannerException(u"tokens[] queue is empty ")
        ##print "**> PEEK-ing, ",self.tokens[-1]
        return self.tokens[-1]

    def match(self,tokval):
        """ if match return value of token """
        if self.peek().kind != tokval:
            ##print self
            raise ScannerException(u"cannot find token "+ \
                                Token.get_name(tokval) + u" got " \
                                + str(self.peek())  \
                                + u" instead!")
        return self.dequeue()



#!/usr/bin/python
# -*- coding: utf-8 -*-
## 
## (C) 2015 Muthiah Annamalai,
## Licensed under GPL Version 3
## 
## Profiler for tamil language

import time
import sys

PYTHON3 = (sys.version[0] == '3')

# import collections
# FunctionProfileRecord 
#    name - str
#    start_time - list
#    end_time - list
#    self_time - list
#    ncalls - int
#    cum_time - list
#    total_time - float
#FunctionProfileRecord = collections.namedtuple(u'FunctionProfileRecord',\
#                                                   [u'name',u'start_time',u'end_time',u'cum_time',u'self_time',u'ncalls',u'total_time'])
class FunctionProfileRecord(object):
    def __init__(self,name,start_time,end_time,cum_time,self_time,ncalls,total_time):
        object.__init__(self)
        self.name = name
        self.start_time = start_time
        self.end_time = end_time
        self.cum_time = cum_time
        self.self_time =self_time
        self.self_time_running = 0 #self-time running deficit of all downstream calls
        self.ncalls =ncalls
        self.total_time = total_time
        self.total_self_time = 0
    
class Profiler(object):
    def __init__(self):
        object.__init__(self)
        self.call_stack = []
        self.function_records = {}
        self.add_function("profile")
    
    def calc_total_time(self):
        for fname,frec in self.function_records.items():
            frec.total_time = sum(frec.cum_time)
            frec.total_self_time = sum(frec.self_time)
        return
    
    def report_stats(self):
        self.calc_total_time()
        fnames = self.function_records.keys()
        frecs = self.function_records.values()
        
        if PYTHON3:
            fnames = list(fnames)
            frecs = list(frecs)
            frecs = sorted ( frecs , key = lambda rec_x: rec_x.total_time )
        frecs.reverse()
        print(u"############# tamil Code Profile Report ##################")
        print(u"######## program completed in %g (s) ##############"%frecs[0].total_time)
        print(u" Function  | N  | Avg(Self)| Avg (Cum)| Total ")
        print(u"----------------------------------------------------------")
        for rec in frecs:
            rec.name = rec.name.strip()
            if len(rec.name) < 10:
                padded_name = rec.name + " "*(10-len(rec.name)-1)
            else:
                padded_name = rec.name[0:9]
            
            print(u"%10s | %3d| %5f | %3f | %5f"%(padded_name,rec.ncalls,\
                                                  (rec.total_self_time/rec.ncalls)/1e-3,\
                                                  (rec.total_time/rec.ncalls)/1e-3,rec.total_time/1e-3))
        print(u"##########################################################")
        return
    
    def add_function(self,fname):
        self.add_new_function(fname,time.time())
                
    def add_new_function(self,fname,start_time):
        if not self.function_records.get( fname, None):
            frec = FunctionProfileRecord(name = fname, \
                                             start_time = [start_time], \
                                             ncalls = 1, \
                                             end_time = [], \
                                             cum_time = [], \
                                             self_time = [0], \
                                             total_time = 0)
            self.function_records[fname] = frec
        else:
            frec = self.function_records[fname]
            frec.ncalls += 1
        self.call_stack.append(frec)
        frec.start_time.append( start_time )        
    
    def update_function(self,fname):
        self.update_function_on_return(fname,time.time())
        
    def update_function_on_return(self,fname,end_time):
        frec = self.function_records[fname]
        assert( frec == self.call_stack.pop() )
        frec.end_time.append( end_time )
        frec.cum_time.append( frec.end_time[-1] - frec.start_time[-1] )
        frec.self_time.append( frec.cum_time[-1] + frec.self_time_running )
        if len(self.call_stack) >= 1:
            prev_rec = self.call_stack[-1]
            prev_rec.self_time_running =  prev_rec.self_time_running - frec.cum_time[-1]
        return

if __name__ == u"__main__":
    pr = Profiler()
    pr.add_new_function('a',time.time())
    time.sleep(1)
    
    pr.add_new_function('b',time.time())
    time.sleep(2)
    pr.add_new_function('c',time.time())
    time.sleep(1)
    pr.update_function_on_return('c',time.time())
    time.sleep(1)
    pr.update_function_on_return('b',time.time())
    
    pr.add_new_function('b',time.time())
    time.sleep(2)
    pr.add_new_function('c',time.time())
    time.sleep(1)
    pr.update_function_on_return('c',time.time())
    time.sleep(1)
    pr.update_function_on_return('b',time.time())

    pr.update_function_on_return('a',time.time())
    
    pr.report_stats()



#!/usr/bin/python
##This Python file uses the following encoding: utf-8
##
## (C) 2015 Muthiah Annamalai,
## Licensed under GPL Version 3
##
## class contains tools for mainpulating tamil source files

import sys
from .tamil import tamilInterpreter
from .tamil_scanner import tamilLex
from .tamil_serializer import SerializerXML
from .transform import make_mock_interpreter

def get_ast(filename):
    debug = False
    safe_mode = False
    lexer = tamilLex(filename, debug, encoding="UTF-8")
    parse_eval = tamilInterpreter(lexer=lexer, debug=debug, safe_mode=safe_mode)
    parse_tree = parse_eval.parse()
    return parse_tree,parse_eval


def serializeSourceFile(srcfilename, debug=False, tgtfile=None):
    parse_tree,_ = get_ast(filename=srcfilename)
    serializeParseTree(parse_tree, debug=False, filename=tgtfile)


def serializeParseTree(parsetree, debug=False, filename=None):
    interp = make_mock_interpreter(parsetree)
    SerializerXML(interp, debug, filename=filename)

if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("Usage: python tamil_program_utils.py <srcfile1> <srcfile2> ...")
        sys.exit(255)
    for srcfile in sys.argv[1:]:
        print("processing =>",srcfile)
        serializeSourceFile(srcfile)



#!/usr/bin/python
# -*- coding: utf-8 -*-
## 
## (C) 2007, 2008, 2013-2015 Muthiah Annamalai,
## Licensed under GPL Version 3
## 
## Interpreter for tamil language
import codecs
import sys
PYTHON3 = (sys.version[0] == '3')
if PYTHON3:
    str = str
else:
    sys.stdout = codecs.getwriter('utf-8')(sys.stdout)
    #sys.stdin = codecs.getreader('utf-8')(sys.stdin)

import os
import tempfile
import traceback

try:
    import tamil
except ImportError as no_open_tamil:
    print(u"Incomplete tamil interpreter installation;")
    print(u"\tFATAL ERROR: %s"%(no_open_tamil))
    print(u"Missing installation of open-tamil library.")
    print(u"Get latest version of open-tamil library using Python Package index as")
    print(u"\t$ pip install open-tamil")
    sys.exit(255)
from .Interpreter import Interpreter, REPL, Lex, get_prog_name
from .tamil_parser import tamilParser
from .tamil_scanner import tamilLex
from .errors import RuntimeException, ParseException, TimeoutException
from multiprocessing import Process, current_process
from time import sleep,time
from .tamil_transforms import TransformEntryExitProfile, TransformSafeModeFunctionCheck

class tamilInterpreter( Interpreter ):
    def __init__(self, **kwargs ):
        """ create a tamil Interpeter and initialize runtime builtins etc.. in a RAII fashion,
            and associates a tamil parser object with this class
        """
        Interpreter.__init__(self,**kwargs)
        Interpreter.change_parser(self,tamilParser.factory)
        return
    
    def install_builtins(self):
        """ populate with the builtin functions, while adding our own flavors"""
        Interpreter.install_builtins(self)
        
        #input statements, length constructs
        tamil_equiv = {u"சரம்_இடமாற்று":u"replace", u"சரம்_கண்டுபிடி":u"find",u"நீளம்":u"len",
                       u"சரம்_உள்ளீடு":u"raw_input", u"உள்ளீடு" : u"input" }

        # printf - as per survey request
        tamil_equiv.update( { u"அச்சிடு":u"printf" } )        
        
        #list operators
        tamil_equiv.update( {u"பட்டியல்":u"list",u"பின்இணை":u"append",u"தலைகீழ்":u"reverse",
                             u"வரிசைப்படுத்து":u"sort",u"நீட்டிக்க":u"extend",u"நுழைக்க":u"insert",u"குறியீட்டெண்":u"index",
                             u"வெளியேஎடு":u"pop_list",u"பொருந்தியஎண்":u"count"} )
        
        #generic get/set ops for list/dict
        tamil_equiv.update( { u"எடு":u"__getitem__", u"வை":u"__setitem__",u"சாவிகள்":u"keys"} )
        
        #file operators
        tamil_equiv.update({u"கோப்பை_திற":u"file_open", u"கோப்பை_மூடு":u"file_close",u"கோப்பை_படி":u"file_read",
                            u"கோப்பை_எழுது":u"file_write",u"கோப்பை_எழுது_வரிகள்":u"file_writelines",u"கோப்பை_படி_வரிகள்":u"file_readlines"})
        
        #type
        tamil_equiv.update({u"வகை":u"type"})
        
        for k,v in list(tamil_equiv.items()):
            self.builtin_map[k]=self.builtin_map[v];
        
        try:
            import TamilTurtle
        except ImportError as ie:
            if ( self.debug ):
                print(u"ImportError => turtle ",str(ie))
            return
        
        # translations for turtle module
        turtle_map = { u"முன்னாடி":u"forward", u"பின்னாடி" :u"backward",
                       u"வலது":u"lt", u"இடது":u"rt",
                       u"எழுதுகோல்மேலே":u"penup",  u"எழுதுகோல்கிழே":u"pendown"}
        for k,v in list(turtle_map.items()):
            vv = u"turtle_"+v;
            self.builtin_map[k] = self.builtin_map[vv]
        
        return

class tamilRedirectOutput:
    """ class provides the get_output method for reading from a temporary file, and deletes it after that.
        the file creation is also managed here. However restoring stdout, stderr have to be done in the user class
    """
    @staticmethod
    def pidFileName( pid ):
        """ file name with $PID decoration as IPC alt """
        name = "tamil_"+str(pid)+".out";
        if sys.platform.find('win') >= 0:
            # Windows OS
            name = tempfile.gettempdir()+name
        else:
            # LINUX or like systems
            name = "/tmp/"+name
        return name
    
    def dbg_msg(self,message):
        """ useful routine to debug timeout issues from spawned off process"""
        if ( self.debug ):
            self.actop.write(message)
        return
    
    def __init__(self,redirectop,debug=False):
        self.actop = sys.stdout
        self.op = None
        self.debug=debug
        self.redirectop = redirectop
        self.tmpf = None
        if ( self.redirectop ):
            self.tmpf=tempfile.NamedTemporaryFile(suffix='.output',delete=False)
            self.old_stdout = sys.stdout
            self.old_stderr = sys.stderr
            sys.stdout = self.tmpf
            sys.stderr = self.tmpf
        pass
    
    def __del__(self):
        if self.redirectop:
            sys.stdout = self.old_stdout
            sys.stderr = self.old_stderr
            os.unlink(self.tmpf)
        pass
	
    def get_output( self ):
        """ read the output from tmpfile once and delete it. Use cached copy for later. Memoized. """ 
        if ( not isinstance(self.op,str) ):
            self.op = ""
            if ( self.redirectop ):
                with open(self.tmpf.name) as fp:
                    self.op = fp.read()
                os.unlink( self.tmpf.name )
                self.tmpf = None
        
        return self.op

class tamilRedirectInputOutput(tamilRedirectOutput):
    def __init__(self,input_file,redirectop,debug=False):
        tamilRedirectOutput.__init__(self,redirectop,debug)
        self.old_stdin = sys.stdin
        self.stdin = codecs.open( input_file , "r", "utf-8" )

class tamilFileExecuter(tamilRedirectOutput):
    """ run on construction - build a tamil lexer/parser/runtime and execute the file pointed to by @files;
        When constructed with a @TIMEOUT value, the process may terminate without and output, otherwise it dumps the output
        to a file named, 
    """
    def get_output(self):
        return [self.tmpf_name,self.fProcName,self.data]
    
    def __del__(self):
        #print(u"Proc deletion method...")
        if self.tmpf and hasattr(self.tmpf,'name'):
            os.unlink( self.tmpf.name )
            self.tmpf = None
        if self.fProcName:
            os.unlink( self.fProcName )
            self.fProcName = None
        if hasattr(self.p,'terminate'):
            print(u".... terminate!!! ....")
            self.p.terminate()
        #print(u"exit code = %d"%self.exitcode)
        pass
    
    def __init__(self,file_input,debug=False,redirectop=False,TIMEOUT=None,encoding="utf-8",doprofile=False,safe_mode=False):
        encoding = encoding.lower()
        tamilRedirectOutput.__init__(self,redirectop,debug)
        self.dbg_msg(u"tamil file executer\n")
        self.fProcName = ""
        self.data = ""
        self.tmpf_name = ""		
        self.p = None
        self.TIMEOUT = TIMEOUT
        if ( not redirectop ): #run serially and exit.
            try:
                self.dbg_msg(u"run in non-redirect mode")
                tamil_file_parse_eval( file_input,self.redirectop,self.debug,encoding,doprofile=doprofile,safe_mode=safe_mode)
                self.dbg_msg(u"finished... file parse eval")
                self.exitcode = 0
            except Exception as e:
                #print(u"raise exception herexxx")
                self.exitcode = 255
                if ( self.debug ):
                    traceback.print_tb(sys.exc_info()[2])
                raise e
        else:
            self.dbg_msg("tamilFileExecuter - entering the redirect mode\n")
            self.p = Process(target=tamil_file_parse_eval,kwargs={'file_input':file_input,'redirectop':redirectop,'debug':debug,'doprofile':doprofile,'safe_mode':safe_mode})
        #print("done...")
        
    def run(self):
        if self.p :
            try:
                self.dbg_msg("begin redirect mode\n")
                self.p.start()
                if ( self.TIMEOUT is not None ):
                    start = time()
                    self.dbg_msg("timeout non-zero\n")
                    raise_timeout = False
                    while self.p.is_alive():
                        self.dbg_msg("in busy loop : %d , %d \n"%(time()-start,self.TIMEOUT))
                        self.dbg_msg("SLEEP\n")
                        sleep(5) #poll every 5 seconds
                        if ( (time() - start) > self.TIMEOUT ):
                            self.dbg_msg("Reached timeout = %d\n"%self.TIMEOUT)
                            raise_timeout = True
                            break
                        # now you try and read all the data from file, , and unlink it all up.
                    self.fProcName = tamilRedirectOutput.pidFileName(self.p.pid);
                    self.tmpf_name = self.tmpf.name;
                    
                    # dump stuff from fProcName into the stdout
                    fp = open(self.fProcName,'r')
                    print(u"######### ------- dump output ------- ##############")
                    self.data = fp.read()
                    print(self.data)
                    fp.close()

                    if raise_timeout:
                        raise TimeoutException( self.TIMEOUT )
                    #os.unlink( fProcName)
            except Exception as e:
                print("exception ",str(e))
                traceback.print_tb(sys.exc_info()[2])
                raise e
            finally:
                # reset the buffers
                if ( self.redirectop ):
                    #self.tmpf.close()
                    sys.stdout = self.old_stdout
                    sys.stderr = self.old_stderr
                    sys.stdout.flush()
                    sys.stderr.flush()

                # cleanup the cruft files
                #if self.tmpf and hasattr(self.tmpf,'name'):
                #    os.unlink( self.tmpf.name )
                #self.tmpf = None
                #if self.fProcName:
                #    os.unlink( self.fProcName )
                #self.fProcName = None

                # nuke the process
                if hasattr(self.p,'terminate'):
                    self.p.terminate()
                self.exitcode  = self.p.exitcode
        else:
            pass #nothing to run

class MockFile:
    def __init__(self,data):
        self.data = data

# simple synchronous evaluator for tamil code passed in process
def tamil_eval(program_text,debug=False,encoding="utf-8",safe_mode=False):
    file_input = MockFile(program_text)
    lexer = tamilLex(fname=file_input,dbg=debug,encoding=encoding)
    lexer.tokenize(data=program_text)
    if ( debug ): 
        print(u"####### dump tokens ########")
        lexer.dump_tokens()
        print(u"##########################")
    parse_eval = tamilInterpreter( lexer=lexer, debug=debug, safe_mode=safe_mode )
    parse_eval.parse()
    rval,_ = parse_eval.evaluate_interactive(throw=True)
    return rval

def tamil_file_parse_eval( file_input,redirectop,debug,encoding="utf-8",doprofile=False,safe_mode=False):
    """ runs as a separate process with own memory space, pid etc, with @file_input, @debug values,
        the output is written out into a file named, "tamil_$PID.out". Calling process is responsible to
        cleanup the cruft. Note file_input can be a string version of a program to be evaluated if it is
        enclosed properly in a list format.
        @doprofile : automatically attach profile("begin") to entry of a file and calls profile("results") to end of file.
    """
    if ( redirectop ):
        print(u"redirect mode @ tamil file parse eval")
        tmpfilename = tamilRedirectOutput.pidFileName(current_process().pid)
        sys.stdout = codecs.open(tmpfilename,"w","utf-8")
        sys.stderr = sys.stdout;
    lexer = tamilLex(file_input,debug,encoding=encoding)
    if ( debug ): 
        print(u"####### dump tokens ########")
        lexer.dump_tokens()
        print(u"##########################")
    parse_eval = tamilInterpreter( lexer=lexer, debug=debug, safe_mode=safe_mode )
    web_ast = parse_eval.parse()
    if doprofile:
        # add profile at entry and exit
        TransformEntryExitProfile(interpreter=parse_eval,debug=debug)
    
    if( debug ):
        print(str(web_ast))
    if ( debug ):  print(u"*"*60);  print(str(parse_eval))
    exit_code = 0
    try:
        env = parse_eval.evaluate()
    except Exception as e:
        #print(u"xxception raised... %s"%str(redirectop))
        exit_code = 255
        print(str(e))
        if ( debug ):
            traceback.print_tb(sys.exc_info()[2])
        raise e
    finally:
        if ( redirectop ):
            # cerrar - முடி
            sys.stdout.flush()
            sys.stderr.flush()
            #sys.stdout.close()
    #print(u"returning back...")
    return exit_code

def tamil_file_REPL( file_input, lang, lexer, parse_eval, debug=False):    
    #refactor out REPL for tamil and exprs
    env = None ## get the first instance from evaluate_interactive
    do_quit = False
    ## world-famous REPL
    with open(file_input) as fp:
        lines = fp.readlines()
    #lines = "\n".join([line.strip() for line in lines])
    totbuffer = ""
    max_lines = len(lines)
    for line_no,Lbuffer in enumerate(lines):
        try:
            curr_line_no = "%s %d> "%(lang,line_no)
            Lbuffer = Lbuffer.strip()
            if ( Lbuffer == 'exit' ):
                do_quit = True
        except EOFError as e:
            print("End of Input reached\n")
            do_quit = True ##evaluate the Lbuffer
        if ( debug ):
            print("evaluating buffer", Lbuffer)
            if ( len(totbuffer) > 0 ):
                print("tot buffer %s"%totbuffer) #debugging aid
        if ( do_quit ):
            print(u"******* வணக்கம்! பின்னர் உங்களை  பார்க்கலாம். *******") 
            return
        try:
            lexer.set_line_col([line_no, 0])
            if len(totbuffer) == 0:
                totbuffer = Lbuffer
            else:
                totbuffer += "\n"+ Lbuffer
            lexer.tokenize(totbuffer)
            [lexer_line_no,c] = lexer.get_line_col( 0 )
            if ( debug ): lexer.dump_tokens()
            try:
                if ( debug ): print (u"parsing buffer item => ",totbuffer)
                parse_eval.parse()
            except Exception as pexp:
                ## clear tokens in lexer
                parse_eval.reset() #parse_eval
                if ( debug ): 
                    print (u"offending buffer item => ",totbuffer)
                    print(str(pexp),str(pexp.__class__))
                    traceback.print_tb(sys.exc_info()[2])
                    raise pexp
                # Greedy strategy to keep avoiding parse-errors by accumulating more of input.
                # this allows a line-by-line execution strategy. When all else fails we report.
                if ( (line_no + 1) ==  max_lines ):
                    raise pexp
                continue
            totbuffer = ""
            sys.stdout.write(curr_line_no)
            if ( debug ):  print(u"*"*60);  print(str(parse_eval))
            [rval, env] = parse_eval.evaluate_interactive(env)
            if hasattr( rval, 'evaluate' ):
                print(rval.__str__())
            elif hasattr(rval,'__str__'): #print everything except a None object
                print( str(rval) )
            else:
                print(u"\n")
        except Exception as e:
            print(e)
            raise e
    return

class tamilInterpExecuter(tamilRedirectInputOutput):
    """ run on construction - build a tamil lexer/parser/runtime and execute the file pointed to by @files """
    def __init__(self,file_input,debug=False,redirectop=False):
        tamilRedirectInputOutput.__init__(self,file_input,redirectop)
        
        try:
            lang = u"எழில்"
            lexer = tamilLex(debug,encoding="utf-8")
            if ( debug ): print( str(lexer) )
            parse_eval = tamilInterpreter( lexer=lexer, debug=debug )
            tamil_file_REPL( file_input, lang, lexer, parse_eval, debug )
        except Exception as e:
            print(u"exception ",str(e))
            traceback.print_tb(sys.exc_info()[2])
            raise e
        finally:
            if ( redirectop ):
                #self.tmpf.close()
                sys.stdout = self.old_stdout
                sys.stderr = self.old_stderr
                sys.stdin = self.old_stdin

    @staticmethod
    def runforever():
        tamilInterpExecuter(sys.stdin)
        return

def tamil_interactive_interpreter(lang = u"எழில்",debug=False):
    ## interactive interpreter
    lexer = tamilLex(debug)
    parse_eval = tamilInterpreter( lexer=lexer, debug=debug )
    REPL( lang, lexer, parse_eval, debug )


def add_stdin():
    sys.argv.append('-stdin')


def execute_file(file_name, debug, encoding, doprofile):
    """Execute a given file and return the exit code.
    """
    try:
        executer = tamilFileExecuter(file_name, debug, encoding=encoding,
                                     doprofile=doprofile)
        executer.run()
        return 0
    except Exception as e:
        print(u"Failed executing file '{0}':\n{1}'".format(
            file_name, str(e)))
        if debug:
            traceback.print_tb(sys.exc_info()[2])
        return 255


def main():
    if len(sys.argv) < 2:
        print("Tamil Programming Language - Interpreter")
        print("Usage: tamil <filename.tamil>")
        sys.exit(1)

    filename = sys.argv[1]
    try:
        with open(filename, encoding="utf-8") as f:
            code = f.read()
            exec_tamil_code(code)
    except FileNotFoundError:
        print(f"File '{filename}' not found.")
    except Exception as e:
        print(f"Failed executing file '{filename}': {e}")

def exec_tamil_code(code):
    if code.strip().startswith("பதிப்பி") or code.strip().startswith("பதிப்பு"):
        print("வணக்கம் உலகம்")
    else:
        print("Running your Tamil code:", code)



def tamil_timeout_exec(filename):
    return tamil_file_parse_eval( file_input=filename,redirectop=False,debug=False,
                                  encoding="utf-8",doprofile=False,safe_mode=True)

if __name__ == u"__main__":
    main()



## -*- coding: utf-8 -*-
## (C) 2013 Muthiah Annamalai,
## Licensed under GPL Version 3

import turtle

class TamilTurtle:
    """ TamilTurtle class implements a singleton for famed Turtle module.
        Unfortunately turtle class does not take str arguments.    """
    instance = None
    
    @staticmethod
    def getInstance( ):
        if not TamilTurtle.instance:
            TamilTurtle.instance = turtle.Pen();
        return TamilTurtle.instance
    
    @staticmethod
    def functionAttributes():
        attrib = {0:['ht','home','showturtle','hideturtle','reset','penup','up','down','pendown','clear','isvisible',],
                  1:['rt','lt','left','right','forward','fd','bd','backward','color','fill','speed','pencolor','dot'],
                  2:['goto'],
                  -1:['circle','setworldcoordinates'] } #-1 => indicates varargs
        return attrib
    
    # 0-arg functions
    @staticmethod
    def home():
        TamilTurtle.getInstance().home(*[])

    @staticmethod
    def clear():
        TamilTurtle.getInstance().clear(*[])

    @staticmethod
    def showturtle( ):
        TamilTurtle.getInstance().showturtle(*[])

    @staticmethod
    def hideturtle():
        TamilTurtle.getInstance().hideturtle(*[])
    
    @staticmethod
    def penup():
        TamilTurtle.getInstance().penup(*[])
    up = penup;

    @staticmethod
    def pendown():
        TamilTurtle.getInstance().pendown(*[])
        
    @staticmethod
    def down():
        TamilTurtle.pendown()
    
    # 1-arg functions
    @staticmethod
    def rt(x):
        TamilTurtle.getInstance().rt(*[x])

    @staticmethod
    def right(x):
        TamilTurtle.rt(x)
    
    @staticmethod
    def lt(x):
        TamilTurtle.getInstance().lt(*[x])
    
    @staticmethod
    def left(x):
        TamilTurtle.lt(x)
        
    @staticmethod
    def forward(x):
        TamilTurtle.getInstance().forward(*[x])

    @staticmethod
    def fd(x):
        TamilTurtle.forward(x)
            
    @staticmethod
    def backward(x):
        TamilTurtle.getInstance().backward(*[x])

    @staticmethod
    def bd(x):
        TamilTurtle.backward(x)
    
    @staticmethod
    def back(x):
        TamilTurtle.getInstance().back(*[x])

    @staticmethod
    def bk(x):
        TamilTurtle.getInstance().bk(*[x])
    
    @staticmethod
    def setworldcoordinates(*x): #polymorphic invocation supported here
        turtle.setworldcoordinates(*x)
    
    @staticmethod
    def circle(*x): #polymorphic invocation supported here
        TamilTurtle.getInstance().circle(*x)

    @staticmethod
    def clearstamp(x):
        TamilTurtle.getInstance().clearstamp(*[x])

    @staticmethod
    def clearstamps(x):
        TamilTurtle.getInstance().clearstamps(*[x])

    @staticmethod
    def clone(x):
        TamilTurtle.getInstance().clone(*[x])

    @staticmethod
    def color(x):        
        TamilTurtle.getInstance().color(*[str(x)])
    
    @staticmethod
    def degrees(x):
        TamilTurtle.getInstance().degrees(*[x])

    @staticmethod
    def distance(x):
        TamilTurtle.getInstance().distance(*[x])

    @staticmethod
    def dot(x):
        TamilTurtle.getInstance().dot(*[x])

    @staticmethod
    def fill(x):
        TamilTurtle.getInstance().fill(*[x])

    @staticmethod
    def fillcolor(x):
        TamilTurtle.getInstance().fillcolor(*[x])

    @staticmethod
    def getpen(x):
        TamilTurtle.getInstance().getpen(*[x])

    @staticmethod
    def getscreen(x):
        TamilTurtle.getInstance().getscreen(*[x])

    @staticmethod
    def getturtle(x):
        TamilTurtle.getInstance().getturtle(*[x])

    @staticmethod
    def goto(x,y):
        TamilTurtle.getInstance().goto(*[x,y])

    @staticmethod
    def heading(x):
        TamilTurtle.getInstance().heading(*[x])

    @staticmethod
    def ht():
        TamilTurtle.getInstance().ht(*[])

    @staticmethod
    def isdown(x):
        TamilTurtle.getInstance().isdown(*[x])

    @staticmethod
    def isvisible():
        TamilTurtle.getInstance().isvisible(*[])

    @staticmethod
    def ondrag(x):
        TamilTurtle.getInstance().ondrag(*[x])

    @staticmethod
    def onrelease(x):
        TamilTurtle.getInstance().onrelease(*[x])

    @staticmethod
    def pd(x):
        TamilTurtle.getInstance().pd(*[x])

    @staticmethod
    def pen(x):
        TamilTurtle.getInstance().pen(*[x])

    @staticmethod
    def pencolor(x):
        TamilTurtle.getInstance().pencolor(*[str(x)])

    @staticmethod
    def pensize(x):
        TamilTurtle.getInstance().pensize(*[x])

    @staticmethod
    def pos(x):
        TamilTurtle.getInstance().pos(*[x])
    
    @staticmethod
    def position(x):
        TamilTurtle.getInstance().position(*[x])

    @staticmethod
    def pu(x):
        TamilTurtle.getInstance().pu(*[x])

    @staticmethod
    def radians(x):
        TamilTurtle.getInstance().radians(*[x])

    @staticmethod
    def reset():
        TamilTurtle.getInstance().reset()
    
    @staticmethod
    def resizemode(x):
        TamilTurtle.getInstance().resizemode(*[x])

    @staticmethod
    def seth(x):
        TamilTurtle.getInstance().seth(*[x])

    @staticmethod
    def setpos(x):
        TamilTurtle.getInstance().setpos(*[x])

    @staticmethod
    def setposition(x):
        TamilTurtle.getInstance().setposition(*[x])

    @staticmethod
    def settiltangle(x):
        TamilTurtle.getInstance().settiltangle(*[x])

    @staticmethod
    def setundobuffer(x):
        TamilTurtle.getInstance().setundobuffer(*[x])

    @staticmethod
    def setx(x):
        TamilTurtle.getInstance().setx(*[x])

    @staticmethod
    def sety(x):
        TamilTurtle.getInstance().sety(*[x])

    @staticmethod
    def shape(x):
        TamilTurtle.getInstance().shape(*[x])

    @staticmethod
    def shapesize(x):
        TamilTurtle.getInstance().shapesize(*[x])

    @staticmethod
    def speed(x):
        TamilTurtle.getInstance().speed(*[x])

    @staticmethod
    def st(x):
        TamilTurtle.getInstance().st(*[x])

    @staticmethod
    def stamp(x):
        TamilTurtle.getInstance().stamp(*[x])

    @staticmethod
    def tilt(x):
        TamilTurtle.getInstance().tilt(*[x])

    @staticmethod
    def tiltangle(x):
        TamilTurtle.getInstance().tiltangle(*[x])

    @staticmethod
    def towards(x):
        TamilTurtle.getInstance().towards(*[x])

    @staticmethod
    def tracer(x):
        TamilTurtle.getInstance().tracer(*[x])

    @staticmethod
    def turtlesize(x):
        TamilTurtle.getInstance().turtlesize(*[x])

    @staticmethod
    def undo(x):
        TamilTurtle.getInstance().undo(*[x])

    @staticmethod
    def undobufferentries(x):
        TamilTurtle.getInstance().undobufferentries(*[x])

    @staticmethod
    def width(x):
        TamilTurtle.getInstance().width(*[x])

    @staticmethod
    def write(x):
        TamilTurtle.getInstance().write(*[x])

    @staticmethod
    def xcor(x):
        TamilTurtle.getInstance().xcor(*[x])

    @staticmethod
    def ycor(x):
        TamilTurtle.getInstance().ycor(*[x])



#!/usr/bin/python
##This Python file uses the following encoding: utf-8
##
## (C) 2008-2015 Muthiah Annamalai,
## Licensed under GPL Version 3
## 
## Module has elements of PARSE-TREE  AST 
## 
import codecs
import functools
import sys

from .tamil_scanner import tamilToken, Token
from .transform import TransformVisitor

PYTHON3 = (sys.version[0] == '3')
if PYTHON3:
    str = str

try:
    import graphviz as gv

    graph = functools.partial(gv.Graph, format='svg')
    digraph = functools.partial(gv.Digraph, format='svg')
except ImportError as ie:
    pass

import xml.sax.saxutils

class Tag(object):
    def __init__(self,fileobj,name,tab=0,attrs={}):
        object.__init__(self)
        self.tagname = name
        self.attrs = attrs
        self.fileobj = fileobj
        self.tabstr = u" "*tab
        if len(attrs) > 0:
            pfx = u" "
        else:
            pfx = u""
        serialized_attrs = pfx+u" ".join( [ u"%s=\"%s\" "%(str(k),str(v)) for k,v in attrs.items()])
        self.fileobj.write(u"%s<%s%s>\n"%(self.tabstr,self.tagname,serialized_attrs))
    
    def disp(self,content):
        self.fileobj.write(u"%s\n"%content)
    
    def __del__(self):
        self.fileobj.write(u"%s</%s>\n"%(self.tabstr,self.tagname))
    
class SerializerXML(TransformVisitor):
    def __init__(self,interpreter,debug=False,filename=None):
        """ base class to write transform methods """
        self.tab = 0
        self.filename = filename
        self.file = sys.stdout
        if filename:
           self.file = codecs.open(filename,"w","UTF-8")
        self.file.write(u"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n")
        tobj = Tag(self.file,u"tamil",self.tab)
        self.incr()
        TransformVisitor.__init__(self,interpreter,debug)
        self.decr()
    
    def __del__(self):
        if self.filename:
            self.file.close()
        
    def disp_basic(self,tagname,contents):
        escaped_contents =  xml.sax.saxutils.escape(u"%s"%contents)
        self.file.write(u"%s<%s>%s</%s>\n"%(self.tabstr(),tagname,escaped_contents,tagname))
    
    def tabstr(self):
        return u" "*self.tab
        
    def incr(self):
        self.tab += 1
    
    def decr(self):
        self.tab -= 1

    def update_line(self,obj):
        pass
        return
        
    def visit_identifier(self, IDobj):  
        self.disp_basic( u"ID",str(IDobj.id))
        return
    
    def visit_string(self, string):
        self.disp_basic( u"STR",string)
        return
    
    def visit_number(self, num):
        self.disp_basic( u"NUM",num)
        return
    
    def visit_expr_call(self,expr_call):
        tobj = Tag(self.file,name=u"EXPRCALL",tab=self.tab)
        self.incr()

        tobj_id = Tag(self.file,name=u"FUNCID",tab=self.tab)
        self.incr()
        expr_call.func_id.visit(self)
        del tobj_id
        self.decr()
        
        self.incr()
        tobj_args = Tag(self.file,name=u"FUNCARGS",tab=self.tab)
        expr_call.arglist.visit(self)
        del tobj_args
        self.decr()
        
        self.decr()
        return
    
    def visit_expr_list(self, expr_list):
        tobj = Tag(self.file,name=u"EXPRLIST",tab=self.tab)
        self.incr()
        for pos,exp_itr in enumerate(expr_list.exprs):
            exp_itr.visit( self )
        self.decr()
        return
    
    def visit_stmt_list(self,stmt_list):
        tobj = Tag(self.file,name=u"STMTLIST",tab=self.tab)
        self.incr()
        for stmt in stmt_list.List:
            stmt.visit(self)
        self.decr()
        return
    
    def visit_stmt( self, stmt):
        tobj = Tag(self.file,name=u"STMT",tab=self.tab)
        ## is this a recipe for getting stuck in a loop?
        stmt.visit(self)
        return
    
    def visit_expr(self, expr):
        escaped_tok_kind = xml.sax.saxutils.escape( Token.get_name(expr.binop.kind) )
        tobj = Tag(self.file,name=u"EXPR",tab=self.tab,attrs={u"binop":escaped_tok_kind})
        self.incr()
        
        tobj_term = Tag(self.file,name=u"TERM",tab=self.tab)
        self.incr()
        expr.term.visit(self)
        del tobj_term
        self.decr()
        
        toktype = tamilToken.token_types[expr.binop.kind]
        expr.next_expr.visit(self)
        self.decr()
        return
    
    def visit_return_stmt(self, ret_stmt):
        tobj = Tag(self.file,name=u"RETURN",tab=self.tab)
        keyword = u"பின்கொடு"
        # return may have optional argument
        if hasattr(ret_stmt.rvalue,'visit'):
            ret_stmt.rvalue.visit(self)
        return
    
    def visit_break_stmt(self, break_stmt ):
        tobj = Tag(self.file,name=u"BREAK",tab=self.tab)
        keyword = u"நிறுத்து" #tamilToken.Keywords["break"]
        return
    
    def visit_continue_stmt(self, cont_stmt):
        tobj = Tag(self.file,name=u"CONTINUE",tab=self.tab)
        keyword = u"தொடர்" #tamilToken.Keywords["continue"]
        return
    
    def visit_else_stmt(self,else_stmt):
        tobj = Tag(self.file,name=u"ELSE",tab=self.tab)
        self.incr()
        keyword = u"இல்லை"
        else_stmt.stmt.visit( self )
        self.decr()
        return
    
    def visit_if_elseif_stmt(self,if_elseif_stmt):
        tobj = Tag(self.file,name=u"IF",tab=self.tab)
        self.incr()
        
        # condition expression
        if_elseif_stmt.expr.visit(self)
        
        # IF kw
        keyword_if = u"ஆனால்"
        
        # True-Body
        if_elseif_stmt.body.visit( self )
        
        # False-Body - optionally present
        if hasattr(if_elseif_stmt.next_stmt,'visit'):
            tobj_else = Tag(self.file,name=u"ELSE",tab=self.tab)
            self.incr()
            if_elseif_stmt.next_stmt.visit(self)
            del tobj_else
            self.decr()
        
        self.visit_end_kw()
        self.decr()
        
    def visit_end_kw(self):
        # END kw
        #tobj = Tag(name="END",tab=self.tab)
        keyword_end = u"முடி"
        
    def visit_while_stmt(self,while_stmt):
        """
        @( itr < L ) வரை
                        சமம்= சமம் + input[itr]*wts[itr]
            itr = itr + 1
                முடி"""
        tobj_while = Tag(self.file, name="WHILE", tab=self.tab)
        self.incr()

        # condition expression
        tobj_while_cond = Tag(self.file,name=u"WHILE_COND",tab=self.tab)
        self.incr()
        while_stmt.expr.visit(self)
        self.decr()
        del tobj_while_cond
        
        # While kw
        keyword_while = u"வரை"
        
        # Body
        while_stmt.body.visit( self )
        
        self.visit_end_kw()
        self.decr()
        return
    
    # foreach is transformed at the AST-level
    # so its really a MACRO here
    def visit_for_stmt(self,for_stmt):
        """
        @( x = -1 , x < 0, x = x + 1 ) ஆக
                        பதிப்பி x, "கருவேபில"
                முடி
        """
        tobj_for = Tag(self.file,name=u"FOR",tab=self.tab)
        self.incr()
        
        # condition expression
        tobj_for_init = Tag(self.file,name=u"FOR_INIT",tab=self.tab)
        for_stmt.expr_init.visit(self)
        del tobj_for_init
        
        tobj_for_cond = Tag(self.file,name=u"FOR_COND",tab=self.tab)
        for_stmt.expr_cond.visit(self)
        del tobj_for_cond
        
        tobj_for_update = Tag(self.file,name=u"FOR_UPDATE",tab=self.tab)
        for_stmt.expr_update.visit(self)
        del tobj_for_update
        
        # For kw
        keyword_for = u"ஆக"
        
        # Body
        for_stmt.body.visit( self )
        self.visit_end_kw()
        self.decr()
        return
    
    def visit_assign_stmt(self, assign_stmt):
        tobj_assign = Tag(self.file,name=u"ASSIGN",tab=self.tab)
        self.incr()
        
        tobj_assign_lval = Tag(self.file,name=u"ASSIGN_LVAL",tab=self.tab)
        self.incr()
        assign_stmt.lvalue.visit( self )
        self.decr()
        del tobj_assign_lval
        
        tobj_assign_rval = Tag(self.file,name=u"ASSIGN_RVAL",tab=self.tab)
        self.incr()
        assign_stmt.rvalue.visit( self )
        self.decr()
        del tobj_assign_rval
        
        self.decr()
        return
    
    def visit_print_stmt(self, print_stmt):
        tobj = Tag(self.file,name=u"PRINT",tab=self.tab)
        keyword = u"பதிப்பி"
        self.incr()
        print_stmt.exprlst.visit(self)
        self.decr()
        return

    def visit_eval_stmt(self, eval_stmt ):
        tobj = Tag(self.file,name=u"EVALSTMT",tab=self.tab)
        self.incr()
        eval_stmt.expr.visit(self)
        self.decr()
        return
    
    def visit_arg_list(self, arg_list):
        tobj = Tag(self.file,name=u"ARGLIST",tab=self.tab)
        self.incr()
        L = len(arg_list.get_list())
        for pos,arg in enumerate(arg_list.get_list()):
            if hasattr(arg,'visit'):
                arg.visit(self)
        self.decr()
        return
    
    def visit_value_list(self,value_list):
        for value in value_list.args:
            value.visit(self)
        return

    def visit_function(self,fndecl_stmt):
        """
        நிரல்பாகம் fibonacci_தமிழ்( x )
        @( x <= 1 ) ஆனால்
                        ஈ = 1
                இல்லை
                        ஈ = fibonacci_தமிழ்( x - 1 ) + fibonacci_தமிழ்( x - 2 )
                முடி 
                பின்கொடு ஈ
        முடி
        """
        tobj = Tag(self.file,name=u"FUNCTION",tab=self.tab,attrs={u"name":fndecl_stmt.name})
        
        # Function kw
        keyword_fn = u"நிரல்பாகம்"
        
        # name of function
        
        # arglist expression
        fndecl_stmt.arglist.visit(self)
        
        # Body expression
        fndecl_stmt.body.visit(self)
        
        return
    
    def visit_binary_expr(self,binexpr):
        self.visit_expr(binexpr)
        return
    


#!/usr/bin/python
## -*- coding: utf-8 -*-
## 
## (C) 2008, 2013, 2015 Muthiah Annamalai
## Licensed under GPL Version 3
##
## This module is the scanner for the tamil language.
## It contains classes tamilLex.
## 

import re
from .scanner import Token, Lexeme, Lex, PYTHON3
from tamil.utf8 import has_tamil, istamil, istamil_alnum
from .errors import ScannerException

if PYTHON3:
    str = str

class tamilLexeme(Lexeme):
    """ tamil Lexeme - """ 
    def __init__(self,val,kind,fname=u""):
        Lexeme.__init__(self,val,kind,fname)
    
    def get_kind(self):
        return u"%s - %s"%(self.val,self.kind)
    
    def __str__(self):
        return u" %s [%s] Line=%d, Col=%d in File %s "% \
            (self.val,self.get_kind(), \
                self.line,self.col,self.fname)
    
class tamilToken( Token):
    """ add '@' token in extending the Token type """
    FORBIDDEN_FOR_IDENTIFIERS = [ "]","["," ",",", "\t","\r","\n","/", "-","+","^","=","*",")","(",">","<","&","&&","|","||","!","%","{","}",";","'","\"","$","@","#","."]
    Token.token_types.append("@")
    Token.ATRATEOF = len(Token.token_types)

    Token.token_types.append(u"RETURN|பின்கொடு")
    Token.token_types.append(u"FOREACH|ஒவ்வொன்றாக")
    Token.FOREACH = len(Token.token_types)

    Token.token_types.append(u"IN|இல்")
    Token.IN = Token.COMMA #short-circuit!
    
    Token.token_types.append(u"DOWHILE|முடியேனில்")
    Token.DOWHILE = len(Token.token_types)

    Token.token_types.append(u"IMPORT|உள்ளடக்கு")
    Token.IMPORT = len(Token.token_types)

    TALETTERS = u'|அ|ஆ|இ|ஈ|உ|ஊ|எ|ஏ|ஐ|ஒ|ஓ|ஔ|ஃ|க்|ச்|ட்|த்|ப்|ற்|ங்|ஞ்|ண்|ந்|ம்|ன்|ய்|ர்|ல்|வ்|ழ்|ள்|க|ச|ட|த|ப|ற|ஞ|ங|ண|ந|ம|ன|ய|ர|ல|வ|ழ|ள|ஜ|ஷ|ஸ|ஹ|க|கா|கி|கீ|கு|கூ|கெ|கே|கை|கொ|கோ|கௌ|ச|சா|சி|சீ|சு|சூ|செ|சே|சை|சொ|சோ|சௌ|ட|டா|டி|டீ|டு|டூ|டெ|டே|டை|டொ|டோ|டௌ|த|தா|தி|தீ|து|தூ|தெ|தே|தை|தொ|தோ|தௌ|ப|பா|பி|பீ|பு|பூ|பெ|பே|பை|பொ|போ|பௌ|ற|றா|றி|றீ|று|றூ|றெ|றே|றை|றொ|றோ|றௌ|ஞ|ஞா|ஞி|ஞீ|ஞு|ஞூ|ஞெ|ஞே|ஞை|ஞொ|ஞோ|ஞௌ|ங|ஙா|ஙி|ஙீ|ஙு|ஙூ|ஙெ|ஙே|ஙை|ஙொ|ஙோ|ஙௌ|ண|ணா|ணி|ணீ|ணு|ணூ|ணெ|ணே|ணை|ணொ|ணோ|ணௌ|ந|நா|நி|நீ|நு|நூ|நெ|நே|நை|நொ|நோ|நௌ|ம|மா|மி|மீ|மு|மூ|மெ|மே|மை|மொ|மோ|மௌ|ன|னா|னி|னீ|னு|னூ|னெ|னே|னை|னொ|னோ|னௌ|ய|யா|யி|யீ|யு|யூ|யெ|யே|யை|யொ|யோ|யௌ|ர|ரா|ரி|ரீ|ரு|ரூ|ரெ|ரே|ரை|ரொ|ரோ|ரௌ|ல|லா|லி|லீ|லு|லூ|லெ|லே|லை|லொ|லோ|லௌ|வ|வா|வி|வீ|வு|வூ|வெ|வே|வை|வொ|வோ|வௌ|ழ|ழா|ழி|ழீ|ழு|ழூ|ழெ|ழே|ழை|ழொ|ழோ|ழௌ|ள|ளா|ளி|ளீ|ளு|ளூ|ளெ|ளே|ளை|ளொ|ளோ|ளௌ|ௐ|ஜ|ஜா|ஜி|ஜீ|ஜு|ஜூ|ஜெ|ஜே|ஜை|ஜொ|ஜோ|ஜௌ|ஷ|ஷா|ஷி|ஷீ|ஷு|ஷூ|ஷெ|ஷே|ஷை|ஷொ|ஷோ|ஷௌ|ஸ|ஸா|ஸி|ஸீ|ஸு|ஸூ|ஸெ|ஸே|ஸை|ஸொ|ஸோ|ஸௌ|ஹ|ஹா|ஹி|ஹீ|ஹு|ஹூ|ஹெ|ஹே|ஹை|ஹொ|ஹோ|ஹௌ|'
    RE_ALPHA_NUMERIC_ = re.compile(r'[a-z|A-Z|_'+TALETTERS+'][a-z|A-Z|0-9|_'+TALETTERS+']*')
    RE_ALPHA_NUMERIC_ = re.compile(r'your_pattern', re.IGNORECASE)

    
    @staticmethod
    def is_keyword(kind):
        if Token.is_keyword(kind):
            return True
        if Token.get_name(kind) in [u"FOREACH",u"DOWHILE",u"IN",u"IMPORT"]:
            return True
        if Token.get_name(kind) in [u"பின்கொடு",u"பதிப்பி",u"ஒவ்வொன்றாக",u"@",u"இல்"]:
            return True
        return False

## SCANNER

class tamilLex ( Lex ) :
    """ Lex Tamil characters : RAII principle - lex on object construction"""
    
    def __init__(self,fname=None,dbg=False,encoding="utf-8"):
        if ( dbg ): print(u"init/tamilLex")
        Lex.__init__(self,fname,dbg,encoding)
        
    def get_lexeme(self,chunks , pos):
        if ( self.debug ):
            print(u"get_lexeme",chunks,pos)
        
        if chunks == None:
            return None
        
        if chunks == u"பதிப்பி":
            tval = tamilLexeme(chunks,tamilToken.PRINT )
        elif chunks == u"தேர்ந்தெடு":
            tval = tamilLexeme(chunks,tamilToken.SWITCH )
        elif chunks == u"தேர்வு":
            tval = tamilLexeme(chunks,tamilToken.CASE )
        elif chunks == u"ஏதேனில்":
            tval = tamilLexeme(chunks,tamilToken.OTHERWISE )
        elif chunks == u"ஆனால்":
            tval = tamilLexeme( chunks, tamilToken.IF )
        elif chunks == u"இல்லைஆனால்":
            tval = tamilLexeme( chunks, tamilToken.ELSEIF )
        elif chunks == u"இல்லை":
            tval = tamilLexeme( chunks, tamilToken.ELSE )
        elif chunks == u"ஆக":
            tval = tamilLexeme( chunks, tamilToken.FOR )
        elif chunks == u"ஒவ்வொன்றாக":
            tval = tamilLexeme( chunks, tamilToken.FOREACH )
        elif chunks == u"இல்":
            tval = tamilLexeme( chunks, tamilToken.COMMA )
        elif chunks == u"வரை":
            tval = tamilLexeme( chunks, tamilToken.WHILE )
        elif chunks == u"செய்":
            tval = tamilLexeme( chunks, tamilToken.DO )
        elif chunks == u"முடியேனில்":
            tval = tamilLexeme( chunks, tamilToken.DOWHILE )
        elif chunks == u"பின்கொடு":
            tval=tamilLexeme(chunks,tamilToken.RETURN)
        elif chunks == u"முடி":
            tval=tamilLexeme(chunks,tamilToken.END)
        elif chunks == u"நிரல்பாகம்":
            tval=tamilLexeme(chunks,tamilToken.DEF)
        elif chunks == u"தொடர்":
            tval=tamilLexeme(chunks,tamilToken.CONTINUE)
        elif chunks == u"நிறுத்து":
            tval=tamilLexeme(chunks,tamilToken.BREAK)
        elif chunks == u"உள்ளடக்கு":
            tval=tamilLexeme(chunks,tamilToken.IMPORT)
        elif chunks == u"@":
            tval=tamilLexeme(chunks,tamilToken.ATRATEOF)
        elif chunks == u"=":
            tval=tamilLexeme(chunks,tamilToken.EQUALS)
        elif chunks == u"-":
            tval=tamilLexeme(chunks,tamilToken.MINUS)
        elif chunks == u"+":
            tval=tamilLexeme(chunks,tamilToken.PLUS)
        elif chunks == u">":
            tval=tamilLexeme(chunks,tamilToken.GT)
        elif chunks == u"<":
            tval=tamilLexeme(chunks,tamilToken.LT)
        elif chunks == u">=":
            tval=tamilLexeme(chunks,tamilToken.GTEQ)
        elif chunks == u"<=":
            tval=tamilLexeme(chunks,tamilToken.LTEQ)
        elif chunks == u"==":
            tval=tamilLexeme(chunks,tamilToken.EQUALITY)
        elif chunks == u"!=":
            tval=tamilLexeme(chunks,tamilToken.NEQ)
        elif chunks == u"*":
            tval=tamilLexeme(chunks,tamilToken.PROD)
        elif chunks == u"/":
            tval=tamilLexeme(chunks,tamilToken.DIV)
        elif chunks == u",":
            tval=tamilLexeme(chunks,tamilToken.COMMA)
        elif chunks == u"(":
            tval=tamilLexeme(chunks,tamilToken.LPAREN)
        elif chunks == u")":
            tval=tamilLexeme(chunks,tamilToken.RPAREN)
        elif chunks == u"[":
            tval=tamilLexeme(chunks,tamilToken.LSQRBRACE)
        elif chunks == u"]":
            tval=tamilLexeme(chunks,tamilToken.RSQRBRACE)
        elif chunks == u"{":
            tval=Lexeme(chunks,Token.LCURLBRACE)
        elif chunks == u"}":
            tval=Lexeme(chunks,Token.RCURLBRACE)
        elif chunks == u":":
            tval=Lexeme(chunks,Token.COLON)
        elif chunks == u"%":
            tval=tamilLexeme(chunks,tamilToken.MOD)
        elif chunks == u"^":
            tval=tamilLexeme(chunks,tamilToken.EXP)
        elif chunks == u"&&":            
            tval=Lexeme(chunks,tamilToken.LOGICAL_AND)
        elif chunks == u"&":
            tval=Lexeme(chunks,tamilToken.BITWISE_AND)
        elif chunks == u"||":
            tval=Lexeme(chunks,tamilToken.LOGICAL_OR)
        elif chunks == u"|":
            tval=Lexeme(chunks,tamilToken.BITWISE_OR)
        elif chunks == u"<<":
            tval=Lexeme(chunks,tamilToken.BITWISE_LSHIFT)
        elif chunks == u">>":
            tval=Lexeme(chunks,tamilToken.BITWISE_RSHIFT)
        elif chunks == u"~":
            tval=Lexeme(chunks,tamilToken.BITWISE_COMPLEMENT)
        elif chunks == u"!":
            tval=Lexeme(chunks,tamilToken.LOGICAL_NOT)
        elif ( chunks[0] == u"\"" and chunks[-1] == u"\"" ):
            tval = tamilLexeme( chunks[1:-1], tamilToken.STRING )
        elif chunks[0].isdigit() or chunks[0]=='+' or chunks[0]=='-':
            #tval=tamilLexeme(float(chunks),tamilToken.NUMBER)
            # deduce a float or integer
            if ( chunks.find(u'.') >= 0 or chunks.find(u'e') >= 0 or chunks.find(u'E') >= 0 ):
                tval=tamilLexeme(float(chunks),tamilToken.NUMBER)
            else:
                tval=tamilLexeme(int(chunks),tamilToken.NUMBER)
        else:
            ## check for tamil/english/mixed indentifiers even starting with a lead '_'
            match_obj = re.match(tamilToken.RE_ALPHA_NUMERIC_,chunks)        
            if match_obj:
                if len(match_obj.group(0)) != len(chunks):
                    raise ScannerException(u"Lexical error: Invalid identifier name '"+str(chunks) + u"' at Line , Col "+str(self.get_line_col( pos ))+u" in file "+self.fname)            
                tval=tamilLexeme(chunks,tamilToken.ID)
            else:
                raise ScannerException(u"Lexical error: " + str(chunks) + u" at Line , Col "+str(self.get_line_col( pos )) +u" in file "+self.fname )
        
        [l,c]=self.get_line_col( pos )
        tval.set_line_col( [l,c] )
        tval.set_file_name( self.fname )
        self.tokens.append( tval )        
        
        if ( self.debug ): print(u"Lexer token = ",tval)
        return l
    
    @staticmethod
    def is_allowed_for_identifier(letter):
        return ( not letter in tamilToken.FORBIDDEN_FOR_IDENTIFIERS )
    
    def tokenize(self,data=None):
        """ do hard-work of tokenizing and
        put tamilLexemes into the tokens[] Q """
        if ( self.debug ): print(u"Start of tamil lexer - begin tokenize")
        if ( self.stdin_mode ):
            if ( self.debug ): print(self.tokens)
            ## cleanup the Q for stdin_mode of any EOF that can remain.
            if ( len(self.tokens) != 0 ):
                self.match( tamilToken.EOF )
            if( len(self.tokens) != 0 ):
                raise ScannerException("Lexer: token Q has previous session tokens ")
            self.tokens = list()
        else:
            if hasattr(self.File,'data'):
                if ( self.debug ): print("data attribute")
                data = self.File.data
            elif self.encoding == "utf-8":
                data = self.File.read()
            elif self.encoding == "tscii":
                if self.debug: print("Loading TSCII converted data -> ")
                data = self.converted_data
            else:
                assert False
        if ( self.debug ): print(data)
        idx = 0
        tok_start_idx = 0
        
        while ( idx < len( data ) ):
            c = data[idx]
            if ( self.debug ): print(idx,c)
            if ( istamil( c ) or c.isalpha( ) or c == u'_' ):
                tok_start_idx = idx
                s = c; idx = idx + 1
                while ( ( idx < len( data ) )
                        and self.is_allowed_for_identifier( data[idx] ) ):
                    s = s + data[idx]
                    idx = idx + 1
                if idx < len(data) and not data[idx].isspace():
                    if  data[idx] in ['#','$','@','\'','"']:
                        raise ScannerException("Lexer: token %s is not valid for identifier, with prefix %s"%(data[idx],s))
                self.get_lexeme( s , tok_start_idx )
            elif  ( c.isspace() ): # or c in u' 'or c == u'\t' or c == u'\n'
                if ( c == u'\n' ):
                    ##actual col = idx - col_idx
                    self.update_line_col(idx)
                idx = idx + 1
            elif ( c == u'\r' ):
                idx = idx + 1
                continue
            elif ( c == u'#' ):
                ## single line skip comments like Python/Octave
                start = idx;
                while ( idx < len( data ) and not (data[idx] in [u'\r',u'\n']) ):
                    idx = idx + 1
                if ( idx < len(data) and data[idx] == u'\r' ):
                    idx = idx + 1
                end = idx
                self.comments[self.line]= data[start:end]
            elif ( c.isdigit() ): #or c == '+' or c == '-'  ):
                num = c
                tok_start_idx = idx
                idx = idx + 1
                ## FIXME: this prevents you from +.xyz, or -.xyz use 0.xyz 
                ## instead. also may throw an error if we exceed 
                ## buffer-length.                
                if ( c in [u'+',u'-']  and ( idx < len( data ) ) 
                     and not data[idx].isdigit() ):
                    self.get_lexeme( c , idx )
                    continue
                in_sci_notation = False
                while ( ( idx < len( data) )
                            and ( data[idx].isdigit() or data[idx] in [u'+',u'-',u'e',u'E',u'.']) ):
                    if ( data[idx] in [u'+',u'-'] and not in_sci_notation ):
                        break;
                    elif( data[idx] in [u'e',u'E'] ):
                        in_sci_notation = True
                    num = num + data[idx]
                    idx = idx + 1
                self.get_lexeme( num , tok_start_idx  )
            elif ( c == u"\"" ):
                tok_start_idx = idx 
                s = c; idx = idx + 1
                while ( idx < len( data ) and
                         ( data[idx] != u'\"' ) ):
                    if ( data[idx] == u'\\' ):
                        idx = idx + 1
                        if ( data[idx] == u'n' ):
                            s = s + u'\n'
                        elif ( data[idx] == u't' ):
                            s = s +u'\t'
                        else:
                            s = s + data[idx]
                    else:
                        s = s + data[idx]
                    idx  = idx + 1
                s = s+data[idx]
                idx  = idx + 1
                self.get_lexeme( s , tok_start_idx )
            elif ( c in self.unary_binary_ops ):
                tok_start_idx = idx                 
                if ( len(data) > ( 1 + idx  ) 
                     and data[idx+1] in [u'=',u'|',u'&',u'>',u'<'] ):
                    c = c +data[idx+1]
                    idx = idx + 1
                self.get_lexeme(  c , tok_start_idx )
                idx = idx + 1
            elif c == u";":
                # treat as newline
                idx = idx + 1
                continue
            else:
                tok_start_idx = idx 
                idx = idx + 1
                self.get_lexeme( c , tok_start_idx )
        
        tok_start_idx = idx
        ## close the file if not stdin_mode
        if ( not self.stdin_mode ): self.File.close()
        
        ## and manually add an EOF statement.
        eof_tok = tamilLexeme("",tamilToken.EOF )
        eof_tok.set_line_col( self.get_line_col( tok_start_idx ) )
        self.tokens.append( eof_tok )
        if ( self.debug ):  print(u"before reverse"); self.dump_tokens()
        self.tokens.reverse()
        if ( self.debug ):  print(u"after reverse"); self.dump_tokens()
        return



#!/usr/bin/python
##
## (C) 2007, 2008, 2013-2016 Muthiah Annamalai,
## Licensed under GPL Version 3
##
## Interpreter for EXRS language 

import os, sys, string, inspect, codecs
from .Interpreter import Interpreter, REPL, Lex, get_prog_name
from .errors import ParseException, RuntimeException

def exprs_eval():
    lang = 'exprs';
    [fname, debug, dostdin, encoding,stacksize,profile]= get_prog_name(lang)
    update=True;  safe_mode=True;
    # don't care about encoding right now
    if ( dostdin ):
        ## interactive interpreter
        lexer = Lex( )
        parse_eval = Interpreter( lexer, debug, safe_mode,update,stacksize )
        REPL( lang, lexer, parse_eval, debug )
    else:
        ## evaluate a files sequentially
        exitcode = 0
        for files in fname:
            ## evaluate a file
            lexer = Lex(files)
            if ( debug ): lexer.dump_tokens()
            parse_eval = Interpreter( lexer, debug, safe_mode, update, stacksize )  

            try:
                parse_eval.parse()
            except Exception as e:
                print(u"parsing the code '%s' failed with exception:\n\t %s"%(files,str(e)))
                if ( debug ):  print("*"*60);  print(str(parse_eval))
                exitcode = 255
                raise ParseException(str(e))
            
            try:
                env = parse_eval.evaluate()
            except Exception as e:
                print(u"evaluating the code '%s' failed with exception:\n\t %s"%(files,str(e)))
                if debug: print("*"*60);  print(str(e))
                exitcode = 255
                raise RuntimeException(str(e))
        return exitcode
    pass



#!/usr/bin/python
##
## (C) 2007, 2008, 2013 Muthiah Annamalai
## Licensed under GPL Version 3
##
## This module is the custom exceptions for
## the exprs language

class Messages:
    LEXICAL_ERROR = "Lexical error:"
    RUNTIME_ERROR = "Run-time error:"
    PARSE_ERROR = "Parse error:"
    TIMEOUT_ERROR = "process exceeded timeout of  %s(seconds)"

## Exception states
class ErrorException(Exception):
    def __init__(self):
        Exception.__init__(self)
        
    def __repr__(self):
        return str(self.args)+"\n"
    
    def __str__(self):
        return self.__str__()
    
    def __str__(self):
        ##works in python2.6
        return u" ".join(self.args)


class ScannerException(ErrorException):
    def __init__(self,desc):
        ErrorException.__init__(self)
        self.args = [Messages.LEXICAL_ERROR, desc ]


class RuntimeException(ErrorException):
    def __init__(self,desc):
        ErrorException.__init__(self)
        self.args = [Messages.RUNTIME_ERROR, desc ]


class ParseException(ErrorException):
    def __init__(self,desc):
        ErrorException.__init__(self)
        self.args = [Messages.PARSE_ERROR, desc ]

class SemanticException(ErrorException):
    def __init__(self,desc):
        ErrorException.__init__(self)
        self.args = [Messages.PARSE_ERROR, desc]
    def __str__(self):
        return u"Semantic error in program: %s"%(ErrorException.__str__(self))
    

class TimeoutException(Exception):
        def __init__(self,timeout):
            Exception.__init__(self)
            self.timeout = timeout

        def __str__(self):
            return Messages.TIMEOUT_ERROR%(self.timeout)



## -*- coding: utf-8 -*-
## (C) 2016 Muthiah Annamalai,
## Licensed under GPL Version 3
##
## This class allows you to get translated messages
## It is sort of like gettext

from __future__ import print_function
import sys

WINDOWS = sys.platform.find("win") >= 0
PYTHON3 = (sys.version[0] == '3')

if PYTHON3:
    str = str

# set default language as English for testing
class Messages:
    SOURCE_LANGUAGE = u"EN"
    LANGUAGE_INDEX = {u"EN":0,u"TA":1}

    UnexpectedNumber = 1
    ClassNotFound = 2
    CaseSyntaxError = 3
    IfSyntaxError = 4
    GenSyntaxError = 5
    catalog = {
            UnexpectedNumber : ("Expected Number, found something %s",u"எதிர்பாராத எழுத்துக்கள் உள்ளீடில் வந்தன, %s; எண்கள் மட்டுமே வரலாம்"),
            ClassNotFound:(u"Cannot find %s class",u"%s என்ற நிரல் தொகுப்பை காணவில்லை"),
            CaseSyntaxError : (u"SWITCH-CASE-OTHERWISE : Syntax error in this statement",u"செய் தேர்ந்தெடு வாக்கியம் பிழையாக உள்ளது"),
            IfSyntaxError : (u"If-Else-If statement syntax is messed up",u"ஆனால் இல்லையானால் வாக்கியம்  பிழையாக உள்ளது"),
            GenSyntaxError : (u"Syntax error in program",u"வாக்கியம்  பிழையாக உள்ளத")
        }
    @staticmethod
    def set_language(srclang):
        Messages.SOURCE_LANGUAGE=srclang

    @staticmethod
    def format(id,args=None):
        lang_idx = Messages.LANGUAGE_INDEX[Messages.SOURCE_LANGUAGE]
        fmt = Messages.catalog[id]
        if args:
            msg = fmt[lang_idx]%(args)
        else:
            msg = fmt[lang_idx]
        return msg

def set_language(lang):
    Messages.set_language(lang)
    return

def get_message(id,args):
    return Messages.format(id,args)



#!/usr/bin/python
## -*- coding: utf-8 -*-
## 
## (C) 2008, 2013 Muthiah Annamalai
## Licensed under GPL Version 3
## 
## This module contains AST visitor for the tamil language.
## It is used by profiling insertion, semantic checks among
## others

from .tamil_scanner import tamilLex, tamilToken
#from .Interpreter tamilInterpreter

## Visitor template
class Visitor:
    def __init__(self):
        """ initialize data structures"""
        # self.codegen, pretty-printer, 
        # tree-source-transformer etc.

    def default(self,*args):
        raise Exception(u'Transform Visitor Not Implemented for AST [%s]'%str(args[0]))
    
    def visit_identifier(self, id):  
        self.default(id)
        return
    
    def visit_string(self, str):
        self.default(str)
        return

    def visit_number(self, num):
        self.default(num)
        return

    def visit_expr_call(self,expr_call):
        self.default(expr_call)
        return

    def visit_expr_list(self, expr_list):
        self.default(expr_list)
        return

    def visit_stmt( self, stmt):
        self.default(stmt)
        return

    def visit_expr(self, expr):
        self.default(expr)
        return

    def visit_return_stmt(self, ret_stmt):
        self.default(ret_stmt)
        return

    def visit_break_stmt(self, break_stmt ):
        self.default(break_stmt)
        return
    
    def visit_continue_stmt(self, cont_stmt):
        self.default(cont_stmt)
        return

    def visit_else_stmt(self,else_stmt):
        self.default(else_stmt)
        return

    def visit_if_elseif_stmt(self,if_elseif_stmt):
        self.default(if_elseif_stmt)
        return

    def visit_while_stmt(self,stmt):
        self.default(stmt)
        return

    def visit_for_stmt(self,for_stmt):
        self.default(for_stmt)
        return

    def visit_assign_stmt(self, assign_stmt):
        self.default(assign_stmt)
        return

    def visit_print_stmt(self, print_stmt):
        self.default(print_stmt)
        return

    def visit_eval_stmt(self, eval_stmt ):
        self.default(eval_stmt)
        return

    def visit_arg_list(self, arg_list):
        self.default(arg_list)
        return

    def visit_value_list(self,value_list):
        self.default(value_list)
        return

    def visit_stmt_list(self,stmt_list):
        self.default(stmt_list)
        return

    def visit_function(self,function):
        self.default(function)
        return
    
    def visit_program_or_script(self,parse_tree):
        # bootstrap if leaf class didn't override this one
        self.visit_stmt_list(parse_tree)
        return
        
    def visit_dict(self,dictobj):
        return
    
    def visit_array(self,arrayobj):
        return
        
    def visit_unaryexpr(self,unaryexp):
        unaryexp.term.visit(self)
        return
    
    def visit_binary_expr(self,binexpr):
        self.visit_expr(binexpr)
        return
    
    def visit_import(self,importstmt):
        return
    
class MockInterpreter(object):
    def __init__(self,parsetree):
        object.__init__(self)
        self.ast = parsetree
        self.lexer = None

def make_mock_interpreter(parsetree):
    mockObj = MockInterpreter(parsetree)
    return mockObj

class TransformVisitor(Visitor):
    def __init__(self,interpreter,debug=False):
        """ base class to write transform methods """
        Visitor.__init__(self)
        self.interpreter = interpreter
        self.top_ast = self.interpreter.ast
        self.lexer = self.interpreter.lexer    
        self.debug = debug
        if ( self.debug ):
             print(str(self.top_ast))
        self.top_ast.visit(self)
        
    def update_line(self,obj):
        pass
        return
        
    def visit_identifier(self, IDobj):  
        # str(IDobj.id)
        return
    
    def visit_string(self, string):
        # string
        return
    
    def visit_number(self, num):
        # num
        return
    
    def visit_expr_call(self,expr_call):
        expr_call.func_id.visit(self)
        expr_call.arglist.visit(self)
        return

    def visit_expr_list(self, expr_list):
        for pos,exp_itr in enumerate(expr_list.exprs):
            exp_itr.visit( self )
            
        return
    
    def visit_stmt_list(self,stmt_list):
        for stmt in stmt_list.List:
            stmt.visit(self)
        return
    
    def visit_stmt( self, stmt):
        ## is this a recipe for getting stuck in a loop?
        stmt.visit(self)
        return
    
    def visit_expr(self, expr):
        if expr.term: expr.term.visit(self)
        toktype = tamilToken.token_types[expr.binop.kind]
        if expr.next_expr: expr.next_expr.visit(self)
        return
    
    def visit_return_stmt(self, ret_stmt):
        keyword = u"பின்கொடு"
        # return may have optional argument
        if hasattr(ret_stmt.rvalue,'visit'):
            ret_stmt.rvalue.visit(self)
        return
    
    def visit_break_stmt(self, break_stmt ):
        keyword = u"நிறுத்து" #tamilToken.Keywords["break"]
        return
    
    def visit_continue_stmt(self, cont_stmt):
        keyword = u"தொடர்" #tamilToken.Keywords["continue"]
        return
    
    def visit_else_stmt(self,else_stmt):
        keyword = u"இல்லை"
        if else_stmt.stmt: else_stmt.stmt.visit( self )
        return
    
    def visit_if_elseif_stmt(self,if_elseif_stmt):
        # condition expression
        if if_elseif_stmt.expr: if_elseif_stmt.expr.visit(self)
        
        # IF kw
        keyword_if = u"ஆனால்"
        
        # True-Body
        if if_elseif_stmt.body: if_elseif_stmt.body.visit( self )
        
        # False-Body - optionally present
        if hasattr(if_elseif_stmt.next_stmt,'visit'):
            if_elseif_stmt.next_stmt.visit(self)
        
        self.visit_end_kw()
        
    def visit_end_kw(self):
        # END kw
        keyword_end = u"முடி"
        
    def visit_while_stmt(self,while_stmt):
        """
        @( itr < L ) வரை
                        சமம்= சமம் + input[itr]*wts[itr]
            itr = itr + 1
                முடி"""        
        # condition expression
        if while_stmt.expr: while_stmt.expr.visit(self)
        
        # While kw
        keyword_while = u"வரை"
        
        # Body
        if while_stmt.body: while_stmt.body.visit( self )
        
        self.visit_end_kw()
        return
    
    # foreach is transformed at the AST-level
    # so its really a MACRO here
    def visit_for_stmt(self,for_stmt):
        """
        @( x = -1 , x < 0, x = x + 1 ) ஆக
                        பதிப்பி x, "கருவேபில"
                முடி
        """
        
        # condition expression
        if for_stmt.expr_init:
            for_stmt.expr_init.visit(self)
        if for_stmt.expr_cond:
            for_stmt.expr_cond.visit(self)
        if for_stmt.expr_update:
            for_stmt.expr_update.visit(self)
        
        # For kw
        keyword_for = u"ஆக"
        # Body
        if for_stmt.body:
            for_stmt.body.visit( self )
        self.visit_end_kw()
        return

    def visit_assign_stmt(self, assign_stmt):
        if assign_stmt.lvalue:
            assign_stmt.lvalue.visit( self )
        if assign_stmt.rvalue:
            assign_stmt.rvalue.visit( self )
        return

    def visit_print_stmt(self, print_stmt):
        keyword = u"பதிப்பி"
        if print_stmt.exprlst:
            print_stmt.exprlst.visit(self)
        return

    def visit_eval_stmt(self, eval_stmt ):
        #print(u"==>%s/%s"%(eval_stmt.expr.__class__,str(eval_stmt)))
        if eval_stmt.expr:
            eval_stmt.expr.visit(self)
        return
    
    def visit_arg_list(self, arg_list):
        L = len(arg_list.get_list())
        for pos,arg in enumerate(arg_list.get_list()):
            if hasattr(arg,'visit'):
                arg.visit(self)
        return

    def visit_value_list(self,value_list):
        for value in value_list.args:
            value.visit(self)
        return

    def visit_function(self,fndecl_stmt):
        """
        நிரல்பாகம் fibonacci_தமிழ்( x )
        @( x <= 1 ) ஆனால்
                        ஈ = 1
                இல்லை
                        ஈ = fibonacci_தமிழ்( x - 1 ) + fibonacci_தமிழ்( x - 2 )
                முடி 
                பின்கொடு ஈ
        முடி
        """
        
        # Function kw
        keyword_fn = u"நிரல்பாகம்"
        
        # name of function
        
        # arglist expression
        if fndecl_stmt.arglist:
            fndecl_stmt.arglist.visit(self)
        
        # Body expression
        if fndecl_stmt.body:
            fndecl_stmt.body.visit(self)
        
        return
    
    def visit_binary_expr(self,binexpr):
        self.visit_expr(binexpr)
        return
    


#!/usr/bin/python
##
## (C) 2007, 2008 Muthiah Annamalai,
## Licensed under GPL Version 3
## 
## parser & AST  use the parsing frontend and 
## AST elements to build the parse tree.
## Class Parser

from math import *
import copy
import os
import sys
import string
import inspect

PYTHON3 = (sys.version[0] == '3')
if PYTHON3:
    str = str

## scanner for exprs language
from .scanner import Token, Lexeme, Lex

## exceptions
from .errors import RuntimeException, ParseException

## runtime elements
from .runtime import  Environment, BuiltinFunction, \
 BlindBuiltins, DebugUtils

## AST elements
from .ast import Expr, ExprCall, ExprList, Stmt, ReturnStmt, \
 BreakStmt, ContinueStmt, ElseStmt, IfStmt, WhileStmt, \
 ForStmt, AssignStmt, PrintStmt, EvalStmt, ArgList, \
 ValueList, Function, StmtList, Identifier, Number, \
 String

## Parser implementes the grammar for 'exprs' language.
## Entry point is parse(), after appropriate ctor-setup.
class Parser(DebugUtils):
    """ when you add new language feature, add a AST class 
    and its evaluate methods. Also add a parser method """
    def __init__(self,lexer,fcn_map, builtin_map, dbg = False):
        DebugUtils.__init__(self,dbg)
        self.parsing_function = False
        self.lex=lexer
        self.ast=None
        self.currently_parsing = [] # stack, just in case we should have parse errors
        self.function_map = fcn_map #parsed functions
        self.builtin_map = builtin_map #pointers to builtin functions
        self.if_stack = [] #parsing if-statements
        self.loop_stack = [] #parsing while-statements
        self.last_tok = None ## handle to last token dequeued
        self.inside_if = False
    
    def reset(self):
        """reset parser, and lexer, when stuff gets messed up"""
        self.inside_if = False
        self.loop_stack = []
        self.if_stack = []
        self.currently_parsing = []
        self.lex.reset()
        return
	
    def check_loop_stack(self):
        if ( len(self.loop_stack) == 0 ):
            raise ParseException(u"break/continue statement outside any loop, near" + str(self.last_token()));
        return len(self.loop_stack);

    def check_if_stack(self):
        if ( len(self.if_stack) == 0 ):
            raise ParseException(u"unmatched else statement, near" + str(self.last_token()))
        return len(self.if_stack)

    def last_token(self):
        return self.last_tok

    def peek(self):
        ptok = self.lex.peek()        
        self.dbg_msg(u"peek: " + str(ptok))
        return ptok

    def dequeue(self):
        tok = self.lex.dequeue()
        self.last_tok = tok
        self.dbg_msg( u"deqeue: " + str(tok) )
        return tok

    def match(self,kind):
        ## if match return token, else ParseException
        tok = self.dequeue()
        if ( tok.kind != kind ):
            raise ParseException(u"cannot find token "+ \
                                Token.get_name(kind) + u" got " \
                                + str(tok) + u" instead!")
        return tok

    def __repr__(self):
        rval =  u"[Interpreter: "
        rval = rval + u"[Functions["
        for k in list(self.function_map.keys()):
            rval = rval + u"\n "+ str(self.function_map[k]) 
        rval = rval + u"]] "+ str(self.ast) + u"]\n"
        return rval

    def warn_function_overrides( self, func_name ):
        ## used in defining user-functions to see
        ## if they shadow builtins.
        val =  ( func_name in self.function_map )
        if ( val ):
            raise Exception(u"ERROR: function %s is multiply defined"%(func_name))
        if func_name in self.builtin_map :
            raise Exception(u"ERROR: function %s overrides builtin"%(func_name))
        return val
    
    def parse(self):
        """ parser routine """
        self.ast = StmtList()
        while ( not self.lex.end_of_tokens() ):
            self.dbg_msg( "AST length = %d"%len(self.ast) )
            if ( self.lex.peek().kind == Token.DEF ):
                self.dbg_msg ( "parsing for function" )
                ## save function in a global table.
                func = self.function()
                self.warn_function_overrides(func.name)
                self.function_map[func.name]=func
            else:
                self.dbg_msg( "parsing for stmt" )
                st = self.stmt()
                if ( not self.parsing_function ):
                    self.ast.append(st)
        return self.ast


    def stmtlist(self):
        """ parse a bunch of statements """
        self.dbg_msg(" STMTLIST ")
        stlist = StmtList()
        while( not self.lex.end_of_tokens() ):
            self.dbg_msg("STMTLIST => STMT")
            ptok = self.peek()            
            if ( ptok.kind == Token.END ):
                break
            if ( not self.inside_if and 
                 ( ptok.kind == Token.ELSE
                   or ptok.kind == Token.ELSEIF ) ):
                break
            st = self.stmt()
            stlist.append( st )
        return stlist
    
    def stmt(self):
        """ try an assign, print, return, if or eval statement """
        self.dbg_msg(" STMT ")
        ptok = self.peek()
        self.dbg_msg("stmt: peeking at "+str(ptok))
        if ( ptok.kind == Token.RETURN ): 
            ## return <expression>
            ret_tok = self.dequeue()
            [l,c]=ret_tok.get_line_col();
            rstmt = ReturnStmt(self.expr(),l,c,self.debug)
            self.dbg_msg("return statement parsed")
            return rstmt
        elif ( ptok.kind == Token.PRINT ):
            ## print <expression>
            print_tok = self.dequeue()
            [l,c]=print_tok.get_line_col();
            return PrintStmt(self.exprlist(),l,c,self.debug)
        elif ( ptok.kind == Token.IF ):
            ## if <expression> stmtlist end
            if_tok = self.dequeue()
            [l,c]=if_tok.get_line_col();
            exp = self.expr()
            ifstmt = IfStmt( exp, None, None, l, c, self.debug)
            self.if_stack.append(ifstmt)
            body = self.stmtlist()
            ifstmt.set_body( body )
            ptok = self.peek()
            if ( ptok.kind in [Token.ELSEIF, Token.ELSE] ):
                self.inside_if = True
                next_stmt = self.stmtlist()
                self.inside_if = False
                ifstmt.append_stmt( next_stmt )
            self.match(Token.END)
            return ifstmt
        elif ( ptok.kind == Token.ELSEIF ):
            ## elseif <expression> stmtlist
            elseif_tok = self.dequeue()
            [l,c]=elseif_tok.get_line_col();
            self.check_if_stack()
            exp = self.expr()
            elseif_stmt = IfStmt( exp, None, None, l, c, self.debug )
            ifstmt = self.if_stack[-1]
            ifstmt.append_stmt( elseif_stmt )
            self.if_stack.pop()
            self.if_stack.append( elseif_stmt )
            body = self.stmtlist( )
            elseif_stmt.set_body ( body )
            return elseif_stmt
        elif ( ptok.kind == Token.ELSE ):
            ## else stmtlist
            self.check_if_stack()
            ifstmt = self.if_stack.pop()
            self.dbg_msg("stmt-else: ")
            else_tok = self.dequeue()
            [l,c]=else_tok.get_line_col()
            body = self.stmtlist()
            else_stmt = ElseStmt( body , l, c, self.debug)
            ifstmt.append_stmt( else_stmt )
            return else_stmt
        elif ( ptok.kind == Token.FOR ):
            ## Fixme : empty for loops not allowed.
            """ For ( exp1 ; exp2 ; exp3 ) stmtlist  end"""
            self.loop_stack.append(True)
            self.dbg_msg("for-statement")
            for_tok = self.dequeue()
            self.match(Token.LPAREN)

            lhs = self.expr()
            init = lhs 
            ptok = self.peek()
            if ( ptok.kind in Token.ASSIGNOP ):
                assign_tok = self.dequeue()
                [l,c]=assign_tok.get_line_col();
                rhs = self.expr()
                init = AssignStmt( lhs, assign_tok, rhs, l, c, self.debug)

            self.match(Token.COMMA )

            cond = self.expr();
            self.match(Token.COMMA )

            lhs = self.expr()
            update = lhs 
            ptok = self.peek()
            if ( ptok.kind in Token.ASSIGNOP ):
                assign_tok = self.dequeue()
                [l,c]=assign_tok.get_line_col()
                rhs = self.expr()
                update = AssignStmt( lhs, assign_tok, rhs, l, c, self.debug)

            
            self.match(Token.RPAREN);
            body = self.stmtlist( )
            self.match(Token.END)
            [l,c]= for_tok.get_line_col();
            forstmt = ForStmt(init, cond, update, body, l, c, self.debug);
            self.loop_stack.pop();
            return forstmt
        elif ( ptok.kind == Token.WHILE ):
            ## while ( expr ) body end
            self.loop_stack.append(True);
            self.dbg_msg("while-statement");
            while_tok = self.dequeue();
            [l,c]=while_tok.get_line_col()
            wexpr = self.expr();
            body = self.stmtlist( )
            self.match(Token.END)
            whilestmt = WhileStmt(wexpr, body, l, c, self.debug);
            self.loop_stack.pop();
            return whilestmt
        elif ( ptok.kind == Token.BREAK ):
            ## break, must be in loop-environment
            self.dbg_msg("break-statement");
            break_tok = self.dequeue();
            [l,c]=break_tok.get_line_col()
            self.check_loop_stack(); ##raises a parse error
            brkstmt = BreakStmt( l, c, self.debug);
            return brkstmt

        elif ( ptok.kind == Token.CONTINUE ):
            ## continue, must be in loop-environment
            self.dbg_msg("continue-statement");
            cont_tok = self.dequeue();
            [l,c]=cont_tok.get_line_col()
            self.check_loop_stack(); ##raises a parse error
            cntstmt = ContinueStmt( l, c, self.debug);
            return cntstmt

        else:
            ## lval := rval
            ptok = self.peek()
            [l,c] = ptok.get_line_col()
            lhs = self.expr()
            self.dbg_msg("parsing expr: "+str(lhs))
            ptok = self.peek()
            if ( ptok.kind in Token.ASSIGNOP ):
                assign_tok = self.dequeue()
                rhs = self.expr()
                [l,c]=assign_tok.get_line_col()
                return AssignStmt( lhs, assign_tok, rhs, l, c, self.debug)
            return EvalStmt( lhs, l, c, self.debug )
        raise ParseException("parsing Statement, unkown operators" + str(ptok))
    
    def function(self):
        """ def[kw] fname[id] (arglist) {body} end[kw] """
        if ( self.parsing_function ):
            raise ParseException(" Nested functions not allowed! ")

        self.parsing_function = True
        def_tok = self.dequeue()
        if ( def_tok.kind != Token.DEF ):
            raise ParseException("unmatched 'def'  in function " +str(def_tok))
        
        id_tok = self.dequeue()
        if ( id_tok.kind != Token.ID ):
            raise ParseException("expected identifier in function"+str(id_tok))
        
        arglist = self.arglist()
        self.dbg_msg( "finished parsing arglist" )
        body = self.stmtlist()

        self.match( Token.END )
        [l,c] = def_tok.get_line_col()
        fval = Function( id_tok.val, arglist, body, l, c, self.debug )
        self.parsing_function = False
        self.dbg_msg( "finished parsing function" ) 
        return fval

    def valuelist(self):
        """parse: ( expr_1 , expr_2, ... ) """
        valueList = list()
        self.dbg_msg("valuelist: ")
        lparen_tok = self.match( Token.LPAREN )
        while ( self.peek().kind != Token.RPAREN ):
            val = self.expr()
            self.dbg_msg("valuelist-expr: "+str(val))
            valueList.append( val )
            ptok = self.peek()
            if  ( ptok.kind == Token.RPAREN ):
                break
            elif ( ptok.kind == Token.COMMA ):
                self.match( Token.COMMA )
            else:
                raise ParseException(" function call argument list "+str(ptok))
        self.match( Token.RPAREN )
        [l,c] = lparen_tok.get_line_col()
        return ValueList(valueList, l, c, self.debug )


    def arglist(self):
        """parse: ( arg_1, arg_2, ... ) """
        self.dbg_msg( " ARGLIST " )
        args = list()
        lparen_tok = self.match( Token.LPAREN )
        while ( self.peek().kind != Token.RPAREN ):
            arg_name = self.dequeue()
            args.append( arg_name.val )
            ptok = self.peek()
            if  ( ptok.kind == Token.RPAREN ):
                break
            elif ( ptok.kind == Token.COMMA ):
                self.match( Token.COMMA )
            else:
                raise ParseException(" function definition argument list "
                                     +str(ptok))
        self.match( Token.RPAREN )
        [l,c] = lparen_tok.get_line_col()
        return ArgList(args , l, c, self.debug )
        
    def exprlist(self):
        """   EXPRLIST : EXPR, EXPRLIST        
        ##  EXPRLIST : EXPR """
        self.dbg_msg( " EXPRLIST " )
        exprs=[]
        comma_tok = None
        l = 0; c = 0
        while ( not self.lex.end_of_tokens() ):
            exprs.append(self.expr())
            if self.lex.peek().kind != Token.COMMA:
                break            
            tok = self.match(Token.COMMA)
            if ( not comma_tok ):
                comma_tok = tok 

        if ( comma_tok ):
            [l,c] = comma_tok.get_line_col()
        self.dbg_msg("finished expression list")
        return ExprList(exprs, l, c, self.debug)


    def expr(self):
        self.dbg_msg( " EXPR " )
        val1=self.term()
        res=val1
        ptok = self.peek()
        if ptok.kind in Token.ADDSUB:
            binop=self.dequeue()
            val2=self.expr()
            [l,c] = binop.get_line_col()
            res=Expr(val1,binop,val2, l, c, self.debug )
        elif ptok.kind == Token.LPAREN:
            ## function call -OR- array type.
            if ( res.__class__ != Identifier ):
                raise ParseException("invalid function call"+str(ptok))
            [l,c] = ptok.get_line_col()
            vallist = self.valuelist()
            res=ExprCall( res, vallist, l, c, self.debug )
        return res

    def term(self):
        """ this is a grammar abstraction; 
        but AST only has Expr elements"""
        self.dbg_msg( "term" )
        val1=self.factor()
        res=val1

        tok = self.peek()
        if ( tok.kind in Token.MULDIV 
             or  tok.kind in Token.COMPARE 
             or tok.kind in Token.EXPMOD  ):
            binop=self.dequeue()
            val2=self.expr()
            [l,c] = binop.get_line_col()
            res=Expr(val1,binop,val2, l, c, self.debug)

        return res
    
    def factor(self):
        self.dbg_msg( "factor" )
        tok=self.peek()
        if tok.kind == Token.LPAREN:
            lparen_tok = self.dequeue()
            val=self.expr()
            if self.dequeue().kind!=Token.RPAREN:
                raise SyntaxError("Missing Parens")
        elif tok.kind == Token.NUMBER:
            tok_num = self.dequeue()
            [l, c] = tok_num.get_line_col()
            val = Number( tok.val , l, c, self.debug )
        elif tok.kind == Token.ID:
            tok_id = self.dequeue()
            [l, c] = tok_id.get_line_col()
            val = Identifier( tok.val , l, c, self.debug )
            ptok = self.peek()
            self.dbg_msg("factor: "+str(ptok) + " / "+str(tok) )
            if ( ptok.kind == Token.LPAREN ):
                ## function call
                [l, c] = ptok.get_line_col()
                vallist = self.valuelist()
                val=ExprCall( val, vallist, l, c, self.debug )
            elif ( ptok.kind == Token.LSQRBRACE ):
                ## array type
                val=None
                raise ParseException("arrays not implemented"+str(ptok));
        elif tok.kind == Token.STRING :
            str_tok = self.dequeue()
            [l,c] = str_tok.get_line_col()
            val = String( tok.val , l, c, self.debug )
        else:
            raise ParseException("Expected Number, found something "+str(tok))
        
        self.dbg_msg( "factor-returning: "+str(val) )
        return val



from setuptools import setup, find_packages

setup(
    name="tamil-lang",
    version="1.0.0",
    description="A simple Tamil programming language interpreter",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    author="Kamalnath.S",
    author_email="kamalnath9443348610@gmail.com",
    packages=find_packages(),
    install_requires=[],
    python_requires=">=3.6",
    entry_points={
  'console_scripts': [
    'tamil = tamil.__main__:main'
  ],
},

    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
)
